import base64
import datetime
import re
import unicodedata
from binascii import Error as BinasciiError
from email.utils import formatdate
from urllib.parse import quote, unquote
from urllib.parse import urlencode as original_urlencode
from urllib.parse import urlparse

from aquilify.utils.datastructures import MultiValueDict
from aquilify.utils.regex_helper import _lazy_re_compile

# Based on RFC 9110 Appendix A.
ETAG_MATCH = _lazy_re_compile(
    r"""
    \A(      # start of string and capture group
    (?:W/)?  # optional weak indicator
    "        # opening quote
    [^"]*    # any sequence of non-quote characters
    "        # end quote
    )\Z      # end of string and capture group
""",
    re.X,
)

MONTHS = "jan feb mar apr may jun jul aug sep oct nov dec".split()
__D = r"(?P<day>[0-9]{2})"
__D2 = r"(?P<day>[ 0-9][0-9])"
__M = r"(?P<mon>\w{3})"
__Y = r"(?P<year>[0-9]{4})"
__Y2 = r"(?P<year>[0-9]{2})"
__T = r"(?P<hour>[0-9]{2}):(?P<min>[0-9]{2}):(?P<sec>[0-9]{2})"
RFC1123_DATE = _lazy_re_compile(r"^\w{3}, %s %s %s %s GMT$" % (__D, __M, __Y, __T))
RFC850_DATE = _lazy_re_compile(r"^\w{6,9}, %s-%s-%s %s GMT$" % (__D, __M, __Y2, __T))
ASCTIME_DATE = _lazy_re_compile(r"^\w{3} %s %s %s %s$" % (__M, __D2, __T, __Y))

RFC3986_GENDELIMS = ":/?#[]@"
RFC3986_SUBDELIMS = "!$&'()*+,;="


def urlencode(query, doseq=False):
    if isinstance(query, MultiValueDict):
        query = query.lists()
    elif hasattr(query, "items"):
        query = query.items()
    query_params = []
    for key, value in query:
        if value is None:
            raise TypeError(
                "Cannot encode None for key '%s' in a query string. Did you "
                "mean to pass an empty string or omit the value?" % key
            )
        elif not doseq or isinstance(value, (str, bytes)):
            query_val = value
        else:
            try:
                itr = iter(value)
            except TypeError:
                query_val = value
            else:
                query_val = []
                for item in itr:
                    if item is None:
                        raise TypeError(
                            "Cannot encode None for key '%s' in a query "
                            "string. Did you mean to pass an empty string or "
                            "omit the value?" % key
                        )
                    elif not isinstance(item, bytes):
                        item = str(item)
                    query_val.append(item)
        query_params.append((key, query_val))
    return original_urlencode(query_params, doseq)


def http_date(epoch_seconds=None):
    return formatdate(epoch_seconds, usegmt=True)


def parse_http_date(date):
    for regex in RFC1123_DATE, RFC850_DATE, ASCTIME_DATE:
        m = regex.match(date)
        if m is not None:
            break
    else:
        raise ValueError("%r is not in a valid HTTP date format" % date)
    try:
        tz = datetime.timezone.utc
        year = int(m["year"])
        if year < 100:
            current_year = datetime.datetime.now(tz=tz).year
            current_century = current_year - (current_year % 100)
            if year - (current_year % 100) > 50:
                year += current_century - 100
            else:
                year += current_century
        month = MONTHS.index(m["mon"].lower()) + 1
        day = int(m["day"])
        hour = int(m["hour"])
        min = int(m["min"])
        sec = int(m["sec"])
        result = datetime.datetime(year, month, day, hour, min, sec, tzinfo=tz)
        return int(result.timestamp())
    except Exception as exc:
        raise ValueError("%r is not a valid date" % date) from exc


def parse_http_date_safe(date):
    try:
        return parse_http_date(date)
    except Exception:
        pass


def base36_to_int(s):
    if len(s) > 13:
        raise ValueError("Base36 input too large")
    return int(s, 36)


def int_to_base36(i):
    char_set = "0123456789abcdefghijklmnopqrstuvwxyz"
    if i < 0:
        raise ValueError("Negative base36 conversion input.")
    if i < 36:
        return char_set[i]
    b36 = ""
    while i != 0:
        i, n = divmod(i, 36)
        b36 = char_set[n] + b36
    return b36


def urlsafe_base64_encode(s):
    return base64.urlsafe_b64encode(s).rstrip(b"\n=").decode("ascii")


def urlsafe_base64_decode(s):
    s = s.encode()
    try:
        return base64.urlsafe_b64decode(s.ljust(len(s) + len(s) % 4, b"="))
    except (LookupError, BinasciiError) as e:
        raise ValueError(e)


def parse_etags(etag_str):
    if etag_str.strip() == "*":
        return ["*"]
    else:
        etag_matches = (ETAG_MATCH.match(etag.strip()) for etag in etag_str.split(","))
        return [match[1] for match in etag_matches if match]


def quote_etag(etag_str):
    if ETAG_MATCH.match(etag_str):
        return etag_str
    else:
        return '"%s"' % etag_str


def is_same_domain(host, pattern):
    if not pattern:
        return False

    pattern = pattern.lower()
    return (
        pattern[0] == "."
        and (host.endswith(pattern) or host == pattern[1:])
        or pattern == host
    )


def url_has_allowed_host_and_scheme(url, allowed_hosts, require_https=False):
    if url is not None:
        url = url.strip()
    if not url:
        return False
    if allowed_hosts is None:
        allowed_hosts = set()
    elif isinstance(allowed_hosts, str):
        allowed_hosts = {allowed_hosts}
    return _url_has_allowed_host_and_scheme(
        url, allowed_hosts, require_https=require_https
    ) and _url_has_allowed_host_and_scheme(
        url.replace("\\", "/"), allowed_hosts, require_https=require_https
    )


def _url_has_allowed_host_and_scheme(url, allowed_hosts, require_https=False):
    if url.startswith("///"):
        return False
    try:
        url_info = urlparse(url)
    except ValueError:  # e.g. invalid IPv6 addresses
        return False
    if not url_info.netloc and url_info.scheme:
        return False
    if unicodedata.category(url[0])[0] == "C":
        return False
    scheme = url_info.scheme
    if not url_info.scheme and url_info.netloc:
        scheme = "http"
    valid_schemes = ["https"] if require_https else ["http", "https"]
    return (not url_info.netloc or url_info.netloc in allowed_hosts) and (
        not scheme or scheme in valid_schemes
    )


def escape_leading_slashes(url):
    if url.startswith("//"):
        url = "/%2F{}".format(url.removeprefix("//"))
    return url


def _parseparam(s):
    while s[:1] == ";":
        s = s[1:]
        end = s.find(";")
        while end > 0 and (s.count('"', 0, end) - s.count('\\"', 0, end)) % 2:
            end = s.find(";", end + 1)
        if end < 0:
            end = len(s)
        f = s[:end]
        yield f.strip()
        s = s[end:]


def parse_header_parameters(line):
    parts = _parseparam(";" + line)
    key = parts.__next__().lower()
    pdict = {}
    for p in parts:
        i = p.find("=")
        if i >= 0:
            has_encoding = False
            name = p[:i].strip().lower()
            if name.endswith("*"):
                name = name[:-1]
                if p.count("'") == 2:
                    has_encoding = True
            value = p[i + 1 :].strip()
            if len(value) >= 2 and value[0] == value[-1] == '"':
                value = value[1:-1]
                value = value.replace("\\\\", "\\").replace('\\"', '"')
            if has_encoding:
                encoding, lang, value = value.split("'")
                value = unquote(value, encoding=encoding)
            pdict[name] = value
    return key, pdict


def content_disposition_header(as_attachment, filename):
    if filename:
        disposition = "attachment" if as_attachment else "inline"
        try:
            filename.encode("ascii")
            file_expr = 'filename="{}"'.format(
                filename.replace("\\", "\\\\").replace('"', r"\"")
            )
        except UnicodeEncodeError:
            file_expr = "filename*=utf-8''{}".format(quote(filename))
        return f"{disposition}; {file_expr}"
    elif as_attachment:
        return "attachment"
    else:
        return None



from typing import List, Tuple

from ..wrappers import Request, Response

class ProxyFix:
    def __init__(
        self,
        app,
        num_proxies: int = 1,
        trusted_proxies: List[str] = None,
        headers_to_check: List[str] = None,
        trusted_hostnames: List[str] = None,
    ):
        self.app = app
        self.num_proxies = num_proxies
        self.trusted_proxies = set(trusted_proxies) if trusted_proxies else set()
        self.headers_to_check = headers_to_check or ['x-forwarded-for', 'x-real-ip', 'forwarded']
        self.trusted_hostnames = set(trusted_hostnames) if trusted_hostnames else set()

    async def __call__(
        self, request: Request, response: Response
    ) -> Response:
        remote_address, forwarded_proto, hostname = self.extract_client_info(request)

        if self.is_trusted_proxy(remote_address):
            request.client = remote_address

        if forwarded_proto:
            request.scope['scheme'] = forwarded_proto

        return response

    def extract_client_info(self, request: Request) -> Tuple[str, str, str]:
        remote_address = request.client
        forwarded_proto = None
        hostname = request.url.hostname if request.url else ''

        for header_name in self.headers_to_check:
            if header_name in request.headers:
                ips = [ip.strip() for ip in request.headers[header_name].split(',')]
                remote_address = self.extract_real_client_ip(ips)
                if header_name.lower() == 'x-forwarded-proto':
                    forwarded_proto = self.extract_forwarded_proto(request.headers[header_name])
                break

        return remote_address, forwarded_proto, hostname

    def extract_real_client_ip(self, ips: List[str]) -> str:
        client_ip = ips[-self.num_proxies]
        if ',' in client_ip:
            client_ip = client_ip.split(',')[-self.num_proxies].strip()

        if client_ip.startswith('[') and client_ip.endswith(']'):
            client_ip = client_ip[1:-1]

        return client_ip.rsplit(':', 1)[0].strip()

    def extract_forwarded_proto(self, header_value: str) -> str:
        proto_values = header_value.split(',')
        for proto in proto_values:
            proto = proto.strip().lower()
            if proto in ('https', 'http'):
                return proto

        return ''

    def is_trusted_proxy(self, remote_address: str) -> bool:
        return remote_address in self.trusted_proxies

    def remove_header_to_check(self, header_name: str) -> None:
        if header_name in self.headers_to_check:
            self.headers_to_check.remove(header_name)



MODLES = """ 
from aquilify.orm.sqlite3 import Table
from aquilify.orm.fields import CharField

# Create your model here...

# Register Models

# Update the `settings` `DATABASE` Dict[Dict[str, str]] -> {
#     MODELS: List[str, str] = [
#         "models.your_model_cls" # pass the path_to model dotted string inside the list.
#     ]
# }
"""






from .column import Column, ColumnType


class TableMeta(type):
    def __getattribute__(cls, item: str) -> ColumnType | object:

        """
        A function to change the class of a column to that.

        :param item: Variable key
        :return: Replaced object or standard object
        """

        data = super().__getattribute__(item)
        if isinstance(data, Column):
            return ColumnType(cls.__tablename__, item, data.type, data.options)

        return data


class Table(object, metaclass=TableMeta):
    __tablename__ = "null"

    def __init__(self, **values) -> None:

        """
        Row initialization for a table with given parameters.

        :param values: The values that the table takes
        """

        self.values = values

    @classmethod
    def columns(cls) -> dict[str, ColumnType]:

        """
        Returns all columns defined in the new table.

        :return: Defined columns
        """

        return {attribute: getattr(cls, attribute)
                for attribute in dir(cls)
                if isinstance(getattr(cls, attribute), ColumnType)}


class DynamicTable(object):
    def __init__(self, name: str, columns: dict[str, Column | ColumnType]) -> None:

        """
        Dynamic table initializer.

        :param name: Table name
        :param columns: Table columns
        """

        self.__tablename__ = name

        for key, column in columns.items():
            self.add(key, column)

    def add(self, key: str, column: Column | ColumnType) -> None:

        """
        Adds a new column to the table.

        :param key: Column name
        :param column: Column object
        :return: Nothing
        """

        self.__setattr__(key, ColumnType(self.__tablename__, key, column.type, column.options))

    def columns(self) -> dict[str, ColumnType]:

        """
        Returns all columns defined in the new table.

        :return: Defined columns
        """

        return {attribute: getattr(self, attribute)
                for attribute in dir(self)
                if isinstance(getattr(self, attribute), ColumnType)}

    def __call__(self, **values) -> Table:

        """
        Row initialization for a dynamic table with given parameters.

        :param values: The values that the table takes
        :return: Initialized table object
        """

        class TempTable(Table):
            __tablename__ = self.__tablename__

        for key, column in self.columns().items():
            setattr(TempTable, key, column)

        return TempTable(**values)



from typing import  Optional
from ..wrappers import Request, Response
import importlib
import logging

logger = logging.getLogger(__name__)

class XFrameOptionsMiddleware:
    ALLOW_DENY_OPTIONS = ["DENY", "SAMEORIGIN"]

    @staticmethod
    def get_allow_from(settings_module: str = 'settings') -> Optional[str]:
        allow_from = None
        try:
            settings = importlib.import_module(settings_module)
            allow_from = getattr(settings, 'X_FAME_OPTIONS', "SAMEORIGIN")
            if allow_from.upper() not in XFrameOptionsMiddleware.ALLOW_DENY_OPTIONS:
                raise ValueError("Invalid 'allow_from' value. Choose from 'DENY' or 'SAMEORIGIN'")
        except ImportError as ie:
            logger.error(f"Settings module '{settings_module}' not found: {ie}")
            raise ImportError(f"Settings module '{settings_module}' not found: {ie}")
        except AttributeError as ae:
            logger.error(f"Attribute 'X_FAME_OPTIONS' not found in '{settings_module}': {ae}")
            raise AttributeError(f"Attribute 'X_FAME_OPTIONS' not found in '{settings_module}': {ae}")
        except ValueError as ve:
            logger.error(str(ve))
            raise ValueError(str(ve))
        except Exception as e:
            logger.error(f"Error fetching 'allow_from' from settings: {e}")
            raise Exception(f"Error fetching 'allow_from' from settings: {e}")
        return allow_from

    async def __call__(self, request: Request, response: Response) -> Response:
        allow_from = self.get_allow_from()
        if allow_from is not None:
            if not 'X-Frame-Options' in response.headers:
                response.headers["X-Frame-Options"] = allow_from
        return response



from typing import Any, Optional
from aquilify.settings import settings
from .storage import default_storage
from aquilify.wrappers import Request, Response

class MessageBeforeStage:
    """
    Middleware class for managing messages before request processing in the Aquilify ASGI framework.

    This middleware assigns message storage to the incoming request object.
    """

    async def __call__(self, request: Request) -> Any:
        """
        Assigns message storage to the incoming request object.

        Args:
        - request (Request): The incoming ASGI HTTP request.

        Returns:
        - Any: The assigned message storage.
        """
        message_storage: Any = default_storage(request)
        request._messages = message_storage
        return message_storage


class MessageMiddleware:
    """
    Middleware class for managing messages during request-response cycle in the Aquilify ASGI framework.

    This middleware updates the response with any stored messages from the request object.
    """

    async def __call__(self, request: Request, response: Response) -> Response:
        """
        Updates the response with any stored messages from the request object.

        Args:
        - request (Request): The incoming ASGI HTTP request.
        - response (Response): The outgoing ASGI HTTP response.

        Returns:
        - Response: The updated ASGI HTTP response.
        """
        if hasattr(request, "_messages"):
            stored_messages: Optional[Any] = request._messages.update(response)
            if stored_messages and settings.DEBUG:
                raise ValueError("Some temporary messages couldn't be stored.")
        return response



import sqlite3

from .filters import MagicFilter
from .column import Column, ColumnType
from .constants import Types
from .exceptions import SessionExecuteError
from .table import Table, DynamicTable
from .utils.dict_factory import dict_factory
from .connection import DatabaseConnectionManager
from .transactions import TransactionContextManager, IsolationLevel

from typing import Callable, Union, Type


class Typing(object):

    """
    Namespace with type hints.
    """

    AnyTable = Union[MagicFilter, DynamicTable, Table, Type[Table]]
    NamespaceTable = Union[DynamicTable, Type[Table]]
    AnyColumn = Union[Column, ColumnType]

class Sqlite3:
    def __init__(self, path: str) -> None:
        self.path = path
        
    def get_path(self) -> str:
        return self.path
    
    def __str__(self) -> str:
        return self.path

class Session(object):
    def __init__(self, tables: list[Typing.NamespaceTable] = None, **kwargs) -> None:

        """
        Creates a new session to work with the database.

        :param path: Path to the database
        :param tables: List of tables to be created during session initialization
        :param kwargs: Other options for opening a database [ More details in `sqlite3.connect(...)` ]
        """
        
        self._connection = DatabaseConnectionManager()._get_connection()
        self._database = sqlite3.connect(self._connection.get_path(), **kwargs)
        self._tables = tables or []

        for table in self._tables:
            self.create(table)

    def create(self, table: Typing.NamespaceTable) -> None:

        """
        Creates a new table in the database.

        :param table: Table or dynamic table
        :return: Nothing
        """

        self._database.execute(f"CREATE TABLE IF NOT EXISTS {table.__tablename__} "
                               f"({', '.join([column.serialize() for column in table.columns().values()])})")
        self._database.commit()

    def clear(self, table: Typing.NamespaceTable) -> None:

        """
        Clears the selected table.

        :param table: Table or dynamic table
        :return: Nothing
        """

        self._database.execute(
            f"DELETE FROM {table.__tablename__}"
        )
        self._database.commit()

    def drop(self, table: Typing.NamespaceTable) -> None:

        """
        Completely removes the table from the database.

        :param table: Table or dynamic table
        :return: Nothing
        """

        self._database.execute(
            f"DROP TABLE IF EXISTS {table.__tablename__}"
        )
        self._database.commit()

    def insert(self, table: Table, replace: bool = False) -> None:

        """
        Adds a new row to the table.

        :param table: Initialized table object
        :param replace: Will replace an existing row
        :return: Nothing
        """

        values = table.values
        
        try:

            self._database.execute(
                f"INSERT {'OR REPLACE' if replace else ''} INTO {table.__tablename__} ({', '.join(values.keys())}) "
                f"VALUES ({', '.join(['?'] * len(values))})", list(values.values())
            )
            self._database.commit()
            return True
        except Exception:
            return False
        
    def update(self, data: Typing.AnyTable, table: Table) -> None:

        """
        Updates the selected rows in the table.

        :param data: Initialized table object
        :param table: Any type of table or magic filter
        :return: Nothing
        """

        if not isinstance(data, (MagicFilter, DynamicTable, Table, type(Table))):
            raise SessionExecuteError("The data is not a successor of MagicFilterData or Table!")

        update = ", ".join([f"{item} = ?" for item in table.values.keys()])

        if isinstance(data, (DynamicTable, Table, type(Table))):
            self._database.execute(
                f"UPDATE {data.__tablename__} SET {update}", (*table.values.values(), )
            )
            return self._database.commit()

        self._database.execute(
            f"UPDATE {data.parameters['table']} SET {update} WHERE {data.query}",
            (*table.values.values(), *data.variables)
        )
        return self._database.commit()

    def delete(self, data: Typing.AnyTable) -> None:

        """
        Removes all rows that match the specified conditions

        :param data: Any type of table or magic filter
        :return: Nothing
        """

        if not isinstance(data, (MagicFilter, DynamicTable, Table, type(Table))):
            raise SessionExecuteError("The data is not a successor of MagicFilterData or Table!")

        if isinstance(data, (DynamicTable, Table, type(Table))):
            self._database.execute(
                f"DELETE FROM {data.__tablename__}"
            )
            return self._database.commit()

        self._database.execute(
            f"DELETE FROM {data.parameters['table']} WHERE {data.query}", data.variables
        )
        return self._database.commit()

    def exists(self, data: Typing.AnyTable) -> bool:

        """
        Checks for the existence of rows satisfying given conditions.

        :param data: Any type of table or magic filter
        :return: Boolean
        """

        if not isinstance(data, (MagicFilter, DynamicTable, Table, type(Table))):
            raise SessionExecuteError("The data is not a successor of MagicFilterData or Table!")

        if isinstance(data, (DynamicTable, Table, type(Table))):
            return not not self._database.execute(
                f"SELECT EXISTS(SELECT * FROM {data.__tablename__})"
            ).fetchone()[-1]

        return not not self._database.execute(
            f"SELECT EXISTS(SELECT * FROM {data.parameters['table']} WHERE {data.query})", data.variables
        ).fetchone()[-1]

    def select(self, data: Typing.AnyTable, items: list[Typing.AnyColumn] = None) -> list[dict[str, object]]:

        """
        Selects certain data from a table that satisfies given conditions.

        :param data: Any type of table or magic filter
        :param items: Elements to select
        :return: List of tuples
        """

        if not isinstance(data, (MagicFilter, DynamicTable, Table, type(Table))):
            raise SessionExecuteError("The data is not a successor of MagicFilterData or Table!")

        select = "*" if not items else ", ".join(
            [f"{item.table}.{item.name}" for item in items]
        )

        if isinstance(data, (DynamicTable, Table, type(Table))):
            return dict_factory(self._database.execute(
                f"SELECT {select} FROM {data.__tablename__}"
            ))

        return dict_factory(self._database.execute(
            f"SELECT {select} FROM {data.parameters['table']} WHERE {data.query}", data.variables
        ))

    def count(self, data: Typing.AnyTable) -> int:

        """
        Counts the number of rows satisfying given conditions.

        :param data: Any type of table or magic filter
        :return: Integer
        """

        if not isinstance(data, (MagicFilter, DynamicTable, Table, type(Table))):
            raise SessionExecuteError("The data is not a successor of MagicFilterData or Table!")

        if isinstance(data, (DynamicTable, Table, type(Table))):
            return self._database.execute(
                f"SELECT COUNT(*) FROM {data.__tablename__}"
            ).fetchone()[-1]

        return self._database.execute(
            f"SELECT COUNT(*) FROM {data.parameters['table']} WHERE {data.query}", data.variables
        ).fetchone()[-1]

    def execute(self, sql: str, parameters: tuple | object = ()) -> sqlite3.Cursor:

        """
        Execute sql query. [ More details in `sqlite3.connect(...)` ]

        :param sql: Sql query
        :param parameters: Query parameters
        :return: SQLite database cursor
        """

        return self._database.execute(sql, parameters)
    
    def execute_many(self, sql: str, parameters: list[tuple] = ()) -> sqlite3.Cursor:
        """
        Execute an SQL query multiple times with different sets of parameters.

        :param sql: SQL query
        :param parameters: List of tuples representing query parameters for multiple executions
        :return: Nothing
        """
        if not isinstance(parameters, list) or not all(isinstance(param, tuple) for param in parameters):
            raise SessionExecuteError("Invalid parameters. Expecting a list of tuples.")

        try:
            with self._database:
                cursor = self._database.cursor()
                cursor.executemany(sql, parameters)
        except sqlite3.Error as e:
            raise SessionExecuteError(f"Error executing multiple queries: {e}")
        
    def fetch_all(self, sql: str, parameters: tuple | object = (), columns: list[str] | None = None) -> dict[str, object] | None:
        """
        Execute an SQL query and fetch all results.

        :param sql: SQL query
        :param parameters: Query parameters
        :param columns: Optional list of column names to fetch
        :return: List of dictionaries representing the rows
        """
        try:
            cursor = self._database.execute(sql, parameters)
            if columns:
                return dict_factory(dict(zip(columns, row)) for row in cursor)
            return dict_factory(cursor)
        except sqlite3.Error as e:
            raise SessionExecuteError(f"Error fetching all rows: {e}")
        
    def transaction(self, isolation_level=IsolationLevel.DEFERRED):
        """Context manager for transactions."""
        return TransactionContextManager(self._database, isolation_level)
    
    def commit(self):
        
        """
        Commits the changes to the database and closes the connection.

        If an error occurs during commit, it rolls back changes and closes the connection.
        """
        
        try:
            self._database.commit()
        except sqlite3.Error:
            self._database.rollback() 
        finally:
            self._database.close()
            
    def close(self) -> None:

        """
        Will close the session.

        :return: Nothing
        """

        self._database.close()

    def __enter__(self) -> "Session":
        return self

    def __exit__(self, _type: object, _value: object, _traceback: object) -> None:
        self.close()


def create_session(**kwargs) -> Callable[[], Session]:

    """
    Creates all tables in the selected database.

    :param path: Path to the database
    :param tables: List of tables to be created during session initialization
    :param kwargs: Other options for opening a database [ More details in `sqlite3.connect(...)` ]
    :return: Local session with given parameters
    """

    with Session(**kwargs) as _:
        return lambda: Session(**kwargs)



# refer @noql -> 5391

from ._templates import (
    Template as Template,
    TemplateError as TemplateError,
    TemplateNotFound as TemplateNotFound,
    TemplateResponse as TemplateResponse
)

from .jinja_template import Jinja2Templates as Jinja2Templates


import importlib

class LifespanConfigError(Exception):
    pass

class ASGILifespanLoader:
    def __init__(self, settings_module='settings'):
        self.settings_module = settings_module
        
    def load_asgi_lifespans(self):
        try:
            settings = importlib.import_module(self.settings_module)
            asgi_lifespans = getattr(settings, 'LIFESPAN_EVENTS', [])
            
            if not asgi_lifespans:
                pass

            if not isinstance(asgi_lifespans, list):
                raise LifespanConfigError("Invalid LIFESPAN_EVENTS configuration")

            lifespans = []

            for lifespans_data in asgi_lifespans:
                module_name, object_name = lifespans_data.get('origin').rsplit('.', 1)
                middleware_module = importlib.import_module(module_name)
                middleware = getattr(middleware_module, object_name)
                if isinstance(middleware, type):
                        middleware = middleware()
                else:
                    middleware
                lifespans.append({'event': lifespans_data.get('event'), 'origin': middleware})

            return lifespans

        except ImportError:
            raise LifespanConfigError(f"Could not import {self.settings_module}.")
        except AttributeError:
            raise LifespanConfigError(f"LIFESPAN_EVENTS not found in {self.settings_module}.")
        except LifespanConfigError as e:
            raise e
        except Exception as e:
            raise LifespanConfigError(f"Error: {e}")


import base64
from aquilify.wrappers import Request, Response
from typing import Optional, Dict, Any, Union

class BasicAuthBeforeStageHandler:
    def __init__(
        self,
        config: Optional[Dict[str, Any]] = None,
        environment: Optional[Dict[str, Any]] = None,
        protected_routes: Optional[Union[list, str]] = ["*"]
    ) -> None:
        self.username = None
        self.password = None
        self.unauthorized_message = "Unauthorized"
        self.auth_realm = 'Basic realm="Access to Aquilify"'
        self.protected_routes = protected_routes if isinstance(protected_routes, list) else [protected_routes]
        
        self._set_credentials(config, environment)
        
    def _set_credentials(self, config: Optional[Dict[str, Any]], environment: Optional[Dict[str, Any]]) -> None:
        if config:
            self.username = config.get("BASIC_AUTH_USERNAME")
            self.password = config.get("BASIC_AUTH_PASSWORD")
            self.unauthorized_message = config.get("UNAUTHORIZED_MESSAGE", "Unauthorized")
            self.auth_realm = config.get("AUTH_REALM", 'Basic realm="Access to Aquilify"')
        elif environment:
            self.username = environment.get("BASIC_AUTH_USERNAME")
            self.password = environment.get("BASIC_AUTH_PASSWORD")
            self.unauthorized_message = environment.get("UNAUTHORIZED_MESSAGE", "Unauthorized")
            self.auth_realm = environment.get("AUTH_REALM", 'Basic realm="Access to Aquilify"')

    @staticmethod
    def _extract_credentials(auth_header: str):
        auth_decoded = base64.b64decode(auth_header.split(' ')[1]).decode('utf-8')
        return tuple(auth_decoded.split(':', 1)) if ':' in auth_decoded else ('', '')

    async def __call__(self, request: Request) -> Response:
        request.scope['user'] = None
        if self._should_apply_middleware(request):
            try:
                auth_header = request.headers.get('Authorization')

                if not auth_header or not auth_header.startswith('Basic '):
                    raise ValueError("Invalid Authorization Header")

                provided_credentials = self._extract_credentials(auth_header)
                if not provided_credentials or provided_credentials != (self.username, self.password):
                    raise ValueError("Invalid Credentials")

                request.scope['user'] = self.username
            except Exception as e:
                request.scope['user'] = None
                return self._unauthorized_response()

    def _should_apply_middleware(self, request: Request) -> bool:
        route = request.url.path
        if "*" in self.protected_routes or route in self.protected_routes:
            return True
        return False

    def _unauthorized_response(self):
        response = Response(self.unauthorized_message, status_code=401)
        response.headers['WWW-Authenticate'] = self.auth_realm
        return response


from functools import wraps
from typing import Union, Callable, Awaitable, List, Dict

from aquilify.exception.base import MethodNotAllowed
from aquilify.wrappers import Request
from aquilify.responses import JsonResponse
from aquilify.core.status import HTTP_405_METHOD_NOT_ALLOWED

import asyncio

ErrorHandler = Union[Callable[[Request], str], Awaitable[Callable[[Request], str]]]
AllowedMethods = Union[List[str], Dict[str, List[str]]]

def require_http_method(methods: AllowedMethods, error_handler: ErrorHandler = None):
    def decorator(func: Callable[..., Awaitable[JsonResponse]]):
        @wraps(func)
        async def wrapper(request: Request, *args, **kwargs) -> JsonResponse:
            allowed_methods = []
            if isinstance(methods, dict):
                allowed_methods = methods.get(request.url.path, [])
            elif isinstance(methods, list):
                allowed_methods = methods
            else:
                raise ValueError("Invalid 'methods' argument provided.")

            if request.method not in allowed_methods:
                error_message = "Method Not Allowed"
                if error_handler:
                    if asyncio.iscoroutinefunction(error_handler):
                        error_message = await error_handler(request)
                    else:
                        error_message = error_handler(request)
                allowed_methods_str = ', '.join(allowed_methods)
                raise MethodNotAllowed(
                    f"{error_message}. Allowed methods: {allowed_methods_str}"
                )
            return await func(request, *args, **kwargs)
        return wrapper
    return decorator


from functools import wraps
from ..wrappers import (
    Request,
    Response
)
from ..exception.base import TooManyRequests, Forbidden
from ..security.csrf import CSRF, CSRFTokenError

_csrf = CSRF()

def csrf_protect(func):
    @wraps(func)
    async def wrapper(request: Request, *args, **kwargs):
        try:
            if request.method in _csrf.protected_methods:
                try:
                    if not await _csrf.validate_csrf_token(request):
                        return Forbidden('CSRF Token Validation Failed')
                except CSRFTokenError:
                    return Forbidden('CSRF Token Validation Failed')

                client_ip = _csrf.get_client_ip(request)

                if _csrf.rate_limiter and not _csrf.rate_limiter.check(client_ip):
                    _csrf.logger.warning(f"Rate limit exceeded for IP {client_ip}")
                    return TooManyRequests("Rate limit exceeded")

                if _csrf.token_refresh_interval and await _csrf.should_refresh_csrf_token(request.cookies.get(_csrf.csrf_token_key)):
                    new_csrf_token = await _csrf.generate_csrf_token(client_ip)
                    return await _csrf.inject_csrf_token(Response(), new_csrf_token)

            return await func(request, *args, **kwargs)
        except Exception as e:
            raise e
    return wrapper


from __future__ import annotations

import hashlib
import hmac
import os
import posixpath
import secrets
from typing import Optional, Tuple

SALT_CHARS: str = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
DEFAULT_PBKDF2_ITERATIONS: int = 600_000

Salt = str
HashedPassword = str
HashMethod = str
HashTuple = Tuple[HashedPassword, HashMethod]
PathComponent = str
SafePath = Optional[str]

_os_alt_seps: list[str] = [
    sep for sep in [os.sep, os.path.altsep] if sep is not None and sep != "/"
]


def gen_salt(length: int) -> Salt:
    if length <= 0:
        raise ValueError("Salt length must be at least 1.")

    return "".join(secrets.choice(SALT_CHARS) for _ in range(length))


def _hash_internal(method: str, salt: str, password: str) -> HashTuple:
    method, *args = method.split(":")
    salt = salt.encode()
    password = password.encode()

    if method == "scrypt":
        if not args:
            n = 2 ** 15
            r = 8
            p = 1
        else:
            try:
                n, r, p = map(int, args)
            except ValueError:
                raise ValueError("'scrypt' takes 3 arguments.") from None

        maxmem = 132 * n * r * p
        return (
            hashlib.scrypt(password, salt=salt, n=n, r=r, p=p, maxmem=maxmem).hex(),
            f"scrypt:{n}:{r}:{p}",
        )
    elif method == "pbkdf2":
        len_args = len(args)

        if len_args == 0:
            hash_name = "sha256"
            iterations = DEFAULT_PBKDF2_ITERATIONS
        elif len_args == 1:
            hash_name = args[0]
            iterations = DEFAULT_PBKDF2_ITERATIONS
        elif len_args == 2:
            hash_name = args[0]
            iterations = int(args[1])
        else:
            raise ValueError("'pbkdf2' takes 2 arguments.")

        return (
            hashlib.pbkdf2_hmac(hash_name, password, salt, iterations).hex(),
            f"pbkdf2:{hash_name}:{iterations}",
        )
    else:
        raise ValueError(f"Invalid hash method '{method}'.")


def hashpw(
    password: str, method: str = "scrypt", salt_length: int = 16
) -> str:
    salt: Salt = gen_salt(salt_length)
    h, actual_method = _hash_internal(method, salt, password)
    return f"{actual_method}${salt}${h}"


def checkpw(pwhash: str, password: str) -> bool:
    try:
        method, salt, hashval = pwhash.split("$", 2)
    except ValueError:
        return False

    return hmac.compare_digest(_hash_internal(method, salt, password)[0], hashval)


def safe_join(directory: str, *pathnames: PathComponent) -> SafePath:
    if not directory:
        directory = "."

    parts = [directory]

    for filename in pathnames:
        if filename != "":
            filename = posixpath.normpath(filename)

        if (
            any(sep in filename for sep in _os_alt_seps)
            or os.path.isabs(filename)
            or filename == ".."
            or filename.startswith("../")
        ):
            return None

        parts.append(filename)

    return posixpath.join(*parts)



import importlib.util

class TimeoutConfigSettings:
    def __init__(self) -> None:
        self.settings_module_path: str = "./settings.py"

    def _fetch_timeout_config(self):
        try:
            settings = self._load_settings_module()
            seconds = getattr(settings, 'TIMEOUT_SECONDS', 10)
            response = getattr(settings, 'TIMEOUT_RESPONSE', "Gateway Timeout")

            return { 'seconds': seconds, 'response': response }
                    
        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings

    def fetch(self):
        data = self._fetch_timeout_config()
        return data


import importlib.util
import os
import re
import ast
from typing import Callable, Awaitable, Optional, List, Union, Any

class BaseSettings:
    def __init__(self) -> None:
        self.settings_path = os.path.join(os.getcwd(), 'settings.py')

    def _settings_env(self):
        if not os.path.exists(self.settings_path):
            raise FileNotFoundError(f"Error: '{self.settings_path}' does not exist")

        try:
            with open(self.settings_path, 'r') as file:
                file_content = file.read()
                comments_removed = re.sub(r'#.*$', '', file_content, flags=re.MULTILINE)
                
                if "use development" not in comments_removed.lower() and "use production" not in comments_removed.lower():
                    raise ValueError(f"Error: Neither 'use development' nor 'use production' found in settings.py. Add 'use development' to your settings.py")
                return True
        except Exception as e:
            raise Exception(f"Error: {e}")

    def compator(self):
        self._settings_env()
        try:
            spec = importlib.util.spec_from_file_location("settings", self.settings_path)
            if spec is None:
                raise ImportError(f"Error: Unable to load '{self.settings_path}'")

            settings = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(settings)

            debug = settings.DEBUG if getattr(settings, 'DEBUG', False) else False
            return debug

        except (FileNotFoundError, ValueError, ImportError, AttributeError, Exception) as e:
            return f"Error: {e}"

class BaseMiddlewareSettings:
    def __init__(self) -> None:
        self.settings_path = os.path.join(os.getcwd(), 'settings.py')
        self.base_dir = os.path.dirname(os.path.abspath(__file__))

    def static_settings(self):
        default_settings = {
            'static_folders': [os.path.join(self.base_dir, 'static')],
            'url_prefix': '/static/',
            'cache_max_age': 0,
            'enable_gzip': True,
            'response_handler': None,
            'chunk_size': 65536
        }

        try:
            spec = importlib.util.spec_from_file_location("settings", self.settings_path)
            if spec is None:
                raise ImportError(f"Error: Unable to load '{self.settings_path}'")

            settings = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(settings)
        except ImportError:
            return default_settings

        static_folders = getattr(settings, "STATICFILE_DIRS", default_settings['static_folders'])
        static_url_prefix = getattr(settings, "STATIC_URL", default_settings['url_prefix'])
        cache_max_age = getattr(settings, "STATIC_MAX_AGE", default_settings['cache_max_age'])
        enable_gzip = getattr(settings, "STATIC_GZIP_COMPRESSION", default_settings['enable_gzip'])
        response_handler = getattr(settings, "STATIC_RESPONSE_HANDLER", default_settings['response_handler'])
        chunk_size = getattr(settings, "STATIC_CHUNK_SIZE", default_settings['chunk_size'])

        return {
            'static_folders': static_folders,
            'url_prefix': static_url_prefix,
            'cache_max_age': cache_max_age,
            'enable_gzip': enable_gzip,
            'response_handler': response_handler,
            'chunk_size': chunk_size
        }

class MiddlewareEntry:
    def __init__(
        self,
        origin: Union[str, dict],
        order: int = 0,
        __init__: str = None,
        conditions: Optional[List[Callable[..., bool]]] = None,
        group: Optional[str] = None,
        active: bool = True,
        excludes: Optional[Callable[..., Awaitable[bool]]] = None,
        type: str = 'http'
    ) -> None:
        self.origin = origin
        self.order = order
        self.__init__ = __init__
        self.conditions = conditions or []
        self.group = group
        self.active = active
        self.excludes = excludes
        self.type = type
        self.middleware = None

def fetchSettingsMiddleware(your_instance):
    settings_module_path = 'settings'
    middlewares = _import_middleware(settings_module_path)
    
    if not middlewares:
        pass

    for middleware_entry in middlewares:
        if middleware_entry.__init__ == "app":
            middleware = _instantiate_middleware(middleware_entry.middleware, your_instance)
        else:
            if middleware_entry.__init__ is not None:
                raise ValueError("Invalid '__init__' value. Expected 'app' but received another value, currently we only support for app instance __init__.")
            middleware = _instantiate_middleware(middleware_entry.middleware)
        _add_middleware_with_params(your_instance, middleware, middleware_entry)

def _import_middleware(settings_module_path: str) -> List[MiddlewareEntry]:
    try:
        settings_module = importlib.import_module(settings_module_path)
        middlewares_list = getattr(settings_module, "MIDDLEWARES", [])
        return [_create_middleware_entry(entry) for entry in middlewares_list]
    except ImportError as e:
        raise ImportError(f"Error importing settings module '{settings_module_path}': {e}")

def _create_middleware_entry(middleware_entry_data) -> MiddlewareEntry:
    if isinstance(middleware_entry_data, str):
        middleware_entry_data = {"origin": middleware_entry_data}
    try:
        middleware_entry = MiddlewareEntry(**middleware_entry_data)
        middleware_entry.middleware = _get_middleware_class(middleware_entry.origin)
        return middleware_entry
    except (ImportError, AttributeError) as e:
        raise ImportError(f"Error importing middleware '{middleware_entry_data.get('origin')}': {e}")

def _get_middleware_class(origin) -> Callable:
    if isinstance(origin, str):
        module_name, class_name = origin.rsplit('.', 1)
        middleware_module = importlib.import_module(module_name)
        return getattr(middleware_module, class_name)
    raise ValueError("Invalid middleware origin format. It should be a string representing the module path.")

def _instantiate_middleware(middleware_class, app=None):
    if callable(getattr(middleware_class, '__call__', None)) and not isinstance(middleware_class, type):
        return middleware_class
    elif isinstance(middleware_class, type):
        if app is not None:
            return middleware_class(app)
        return middleware_class()
    else:
        raise ValueError("Invalid middleware format. It should be either a function or a class.")

def _add_middleware_with_params(instance, middleware, middleware_entry):
    instance.add_middleware(
        middleware=middleware,
        order=middleware_entry.order,
        conditions=middleware_entry.conditions,
        group=middleware_entry.group,
        active=middleware_entry.active,
        excludes=middleware_entry.excludes,
        type=middleware_entry.type
    )

class StageHandler:
    def _exclude_comments(self, code: str) -> str:
        """Exclude comments from the code."""
        parsed_code = ast.parse(code)
        modified_code = ast.unparse(parsed_code)
        return modified_code

    def load_settings(self):
        try:
            with open("./settings.py", "r") as file:
                code = file.read()
                code = self._exclude_comments(code)
                
                spec = importlib.util.spec_from_loader("settings", loader=None, origin="./settings.py")
                settings = importlib.util.module_from_spec(spec)
                
                settings.__file__ = "./settings.py"  # Define __file__ manually
                
                exec(code, settings.__dict__)
                return settings
        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def load_middleware_from_path(self, middleware_path: str) -> Any:
        try:
            module_path, class_name = middleware_path.rsplit('.', 1)
            module = importlib.import_module(module_path)
            middleware_class = getattr(module, class_name)

            if isinstance(middleware_class, type):
                try:
                    middleware_instance = middleware_class()
                    return middleware_instance
                except Exception as e:
                    raise ImportError(f"Error instantiating middleware class: {e}")
            else:
                return middleware_class 

        except (ModuleNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading middleware: {e}")

    def process_stage_handlers(self, instance):
        settings = self.load_settings()
        STAGE_HANDLERS = getattr(settings, 'STAGE_HANDLERS', None)
        
        if STAGE_HANDLERS is None:
            raise AttributeError("STAGE_HANDLERS not found in settings.py")

        for handler in STAGE_HANDLERS:
            middleware_path = handler.get('origin')
            stage = handler.get('stage')

            if stage not in ['before', 'after']:
                raise ValueError(f"Invalid stage '{stage}'. Stage must be either 'before' or 'after'.")

            try:
                middleware_class = self.load_middleware_from_path(middleware_path)

                instance.request_stage_handlers.setdefault(stage, []).append(
                    (
                        middleware_class,
                        handler.get('order', 0),
                        handler.get('conditions', None)
                    )
                )

                if handler.get('group'):
                    instance.grouped_request_stages \
                        .setdefault(handler.get('group'), {}) \
                        .setdefault(stage, []).append(middleware_class)

                if handler.get('exclude'):
                    instance.excluded_stages.setdefault(handler.get('exclude'), set()).add(middleware_class)

                if handler.get('inherit'):
                    instance._inherit_from_group(stage, handler.get('group'), handler.get('inherit'))

            except ImportError as e:
                raise ImportError(f"Error processing middleware: {e}")


import os
import email.parser
import mimetypes
from .filestorage import FileStorage

class FormParser:
    @staticmethod
    async def parse(body, boundary):
        form_data = {}
        parser = email.parser.BytesParser()

        parts = body.split(b'--' + boundary.encode())
        for part in parts:
            part = part.strip()
            if not part:
                continue

            msg = parser.parsebytes(part)

            if 'Content-Disposition' in msg:
                content_disposition = msg['Content-Disposition']
                field_name, file_name = FormParser.parse_content_disposition(content_disposition)

                if file_name:
                    # This is a file upload
                    file_data = {
                        'filename': file_name,
                        'content_type': msg.get_content_type(),
                        'data': msg.get_payload(decode=True)
                    }
                    form_data[field_name] = await FileStorage.create(**file_data)
                elif field_name:
                    # This is a regular form field
                    field_value = msg.get_payload(decode=True).decode('utf-8')
                    form_data[field_name] = field_value

        return form_data

    @staticmethod
    def parse_content_disposition(content_disposition):
        field_name = None
        file_name = None

        disposition_params = content_disposition.split(';')
        for param in disposition_params:
            param = param.strip()
            if param.startswith('name='):
                field_name = param.split('=')[1].strip('"')
            elif param.startswith('filename='):
                file_name = param.split('=')[1].strip('"')

        return field_name, file_name

    @staticmethod
    def parse_query_string(query_string):
        query_data = {}
        if query_string:
            query_params = query_string.split('&')
            for param in query_params:
                param = param.split('=')
                if len(param) == 2:
                    key, value = param
                    query_data[key] = value
        return query_data
    
    @staticmethod
    def parseUrlEncoded(body):
        form_data = {}
        key_value_pairs = body.split('&')
        for pair in key_value_pairs:
            key, value = pair.split('=')
            form_data[key] = value
        return form_data


from .schema import (
    Schema as Schema,
    Field as Field
)

from .exceptions import (
    ValidationError as ValidationError,
    FieldValidationError as FieldValidationError,
    FieldLengthError as FieldLengthError,
    FieldNotAllowedError as FieldNotAllowedError,
    FieldRequiredError as FieldRequiredError,
    FieldValueError as FieldValueError,
    SchemaLoadError as SchemaLoadError,
)

from . import fields as fields

# Inline Fields mocked due to circulam... PEP-383

"""
from ._fields import (
    String as String,
    Integer as Integer,
    IPAddress as IPAddress,
    Bool as Bool,
    Email as Email,
    Date as Date,
    Decimal as Decimal,
    URL as URL,
    UUID as UUID,
    File as File,
    Float as Float,
    ListField as List,
    NestedSchema as Nested,
    Age as Age,
    Dictonary as Dict,
    And as And,
    Password as Password,
    Length as Length,
    Range as Range,
    Regexp as Regexp,
    Equal as Equal,
    Field as Field
)

"""


import importlib.util

class CSPConfigSettings:
    def __init__(self) -> None:
        self.settings_module_path: str = "./settings.py"

    def _fetch_csp_config(self):
        try:
            settings = self._load_settings_module()

            csp_directives = getattr(settings, "CSP_DIRECTIVES", { 'default-src': ["'self'"] })
            report_uri = getattr(settings, "CSP_REPORT_URI", '')
            report_only = getattr(settings, "CSP_REPORT_ONLY", False)
            enable_violation_handling = getattr(settings, "CSP_ENABLE_VIOLATION_HANDLING", False)
            report_sample_weight = getattr(settings, "CSP_REPORT_SAMPLE_WEIGHT", 0.0)
            violation_report_endpoint = getattr(settings, "CSP_VIOLATION_REPORT_ENDPOINT", '/violation-report')
            log_file_path = getattr(settings, "CSP_LOG_FILE_PATH", 'csp_report.log')
            security_headers = getattr(settings, "CSP_SECURITY_HEADERS", None)
            nonce_length = getattr(settings, "CSP_NONCE_LENGTH", 16)
            nonce_algorithm = getattr(settings, "CSP_NONCE_ALGORITHM", 'sha256')
            referrer_policy = getattr(settings, "CSP_REFERRER_POLICY", 'strict-origin-when-cross-origin')
            hsts_max_age = getattr(settings, "CSP_HSTS_MAX_AGE", 31536000)
            hsts_include_subdomains = getattr(settings, "CSP_HSTS_INCLUDE_SUBDOMAINS", True)
            hsts_preload = getattr(settings, "CSP_HSTS_PRELOAD", False)
            feature_policy = getattr(settings, "CSP_FEATURE_POLICY", None)
            x_content_type_options = getattr(settings, "CSP_X_CONTENT_TYPE_OPTIONS", 'nosniff')
            x_frame_options = getattr(settings, "CSP_X_FRAME_OPTIONS", 'SAMEORIGIN')
            x_xss_protection = getattr(settings, "CSP_X_XSS_PROTECTION", '1; mode=block')
            expect_ct = getattr(settings, "CSP_EXPECT_CT", None)
            cross_origin_opener_policy = getattr(settings, "CSP_CROSS_ORIGIN_OPENER_POLICY", None)
            cross_origin_embedder_policy = getattr(settings, "CSP_CROSS_ORIGIN_EMBEDDER_POLICY", None)
            force_https = getattr(settings, "CSP_FORCE_HTTPS", False)
            referrer_policy_feature = getattr(settings, "CSP_REFERRER_POLICY_FEATURE", False)
            referrer_policy_no_referer = getattr(settings, "CSP_REFERRER_POLICY_NO_REFERER", False)
            referrer_policy_no_referrer_when_downgrade = getattr(settings, "CSP_REFERRER_POLICY_NO_REFERRER_WHEN_DOWNGRADE", False)

            csp_settings = {
                "csp_directives": csp_directives,
                "report_uri": report_uri,
                "report_only": report_only,
                "enable_violation_handling": enable_violation_handling,
                "report_sample_weight": report_sample_weight,
                "violation_report_endpoint": violation_report_endpoint,
                "log_file_path": log_file_path,
                "security_headers": security_headers,
                "nonce_length": nonce_length,
                "nonce_algorithm": nonce_algorithm,
                "referrer_policy": referrer_policy,
                "hsts_max_age": hsts_max_age,
                "hsts_include_subdomains": hsts_include_subdomains,
                "hsts_preload": hsts_preload,
                "feature_policy": feature_policy,
                "x_content_type_options": x_content_type_options,
                "x_frame_options": x_frame_options,
                "x_xss_protection": x_xss_protection,
                "expect_ct": expect_ct,
                "cross_origin_opener_policy": cross_origin_opener_policy,
                "cross_origin_embedder_policy": cross_origin_embedder_policy,
                "force_https": force_https,
                "referrer_policy_feature": referrer_policy_feature,
                "referrer_policy_no_referer": referrer_policy_no_referer,
                "referrer_policy_no_referrer_when_downgrade": referrer_policy_no_referrer_when_downgrade
            }
            return csp_settings

        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings
    
    def fetch(self):
        data = self._fetch_csp_config()
        return data





# Aquilify devops / Test _app @noql :: 3211

from __future__ import annotations

import asyncio
import re
import zlib
import json
from collections import defaultdict

from ..wrappers import (
    Request,
    Response
)

from typing import (
    Callable,
    Any,
    List,
    Type,
    Dict,
    Optional,
    TypeVar,
    Tuple,
    Awaitable,
    Pattern,
)

from ..exception.base import (
    BadRequest,
    NotFound,
    MethodNotAllowed,
    Unauthorized,
    Forbidden,
    InternalServerError
)

from ..exception.__handler import handle_exception

T = TypeVar("T")

class Aquilify:
    def __init__(self, debug: bool = False) -> None:
        self.routes: List[
            Tuple[
                str,
                List[str],
                Callable[..., Awaitable[T]],
                List[str],
                Pattern,
                Type[T],
                str,
                Dict[str, Any],
            ]
        ] = []
        self.exception_handlers: Dict[
            Type[Exception], Callable[..., Awaitable[T]]
        ] = {}
        self.middlewares: List[Callable[..., Awaitable[T]]] = []
        self.middleware_order: List[Tuple[Callable[..., Awaitable[T]], int]] = []
        self.url_patterns: Dict[str, str] = {}
        self.request_transformers: List[Callable[..., Awaitable[Request]]] = []
        self.response_transformers: List[
            Callable[..., Awaitable[Response]]
        ] = []
        self.compression_enabled: bool = False
        self.messages: List[Callable[..., Awaitable[T]]] = []
        self.middleware_groups: Dict[str, List[Callable[..., Awaitable[T]]]] = {}
        self.middleware_activation: Dict[Callable[..., Awaitable[T]], bool] = {}
        self.middleware_dependencies: Dict[Callable[..., Awaitable[T]], List[Callable[..., Awaitable[T]]]] = defaultdict(list)
        self.middleware_exclusions: Dict[Callable[..., Awaitable[T]], List[Callable[..., Awaitable[T]]]] = defaultdict(list)
        self.debug = debug

    def add_exception_handler(
        self, exception_cls: Type[Exception], handler: Callable[..., Awaitable[T]]
    ) -> None:
        self.exception_handlers[exception_cls] = handler

    def route(
        self,
        path: str,
        methods: Optional[List[str]] = None,
        response_model: Optional[Type[T]] = None,
        endpoint: Optional[str] = None,
        strict_slashes: bool = True,
    ) -> Callable[..., Awaitable[T]]:
        allowed_methods = ["GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS", "PATCH", "TRACE"]
        try:
            if methods is not None:
                for method in methods:
                    if method not in allowed_methods:
                        raise ValueError(f"Invalid HTTP method provided: {method}")

            if methods is None:
                methods = ["GET"]

            def decorator(
                handler: Callable[..., Awaitable[T]]
            ) -> Callable[..., Awaitable[T]]:
                converted_path, path_regex = self._regex_coverter(
                    path, strict_slashes
                )
                self.routes.append(
                    (
                        converted_path,
                        methods,
                        handler,
                        path_regex,
                        response_model,
                        endpoint,
                    )
                )
                return handler

            return decorator
        except Exception as e:
            if self.debug:
                print(e)
            else:
                InternalServerError()

    def _regex_coverter(
        self, path: str, strict_slashes: bool
    ) -> Tuple[str, Pattern]:
        if strict_slashes:
            pattern = re.sub(r"{([^}]+)}", r"(?P<\1>[^/]+)", path)
        else:
            pattern = re.sub(r"{([^}]+)}", r"(?P<\1>[^/]+/?)", path)
        pattern = f"^{pattern}$"
        path_regex = re.compile(pattern)
        return pattern, path_regex
    
    def middleware(
        self,
        middleware: Callable[..., Awaitable[T]],
        order: Optional[int] = 0,
        conditions: Optional[List[Callable[..., bool]]] = None,
        group: Optional[str] = None,
        active: bool = True,
        excludes: Optional[Callable[..., Awaitable[T]]] = None,
    ) -> None:
        middleware_entry = {
            "middleware": middleware,
            "order": order,
            "conditions": conditions,
            "group": group,
            "excludes": excludes,
        }
        self.middlewares.append(middleware_entry)
        self.middlewares.sort(key=lambda m: m["order"])
        if group:
            self.middleware_groups[group].append(middleware)
        self.middleware_activation[middleware] = active
        if excludes:
            self.middleware_exclusions[excludes].append(middleware)

    async def apply_middlewares(
        self, request: Request, response: Response
    ) -> Response:
        executed_middlewares = set()

        for middleware_entry in self.middlewares:
            middleware = middleware_entry["middleware"]
            conditions = middleware_entry.get("conditions")
            group = middleware_entry.get("group")
            excludes = middleware_entry.get("excludes")

            if group and middleware not in self.middleware_groups[group]:
                continue

            if not self.middleware_activation.get(middleware, True):
                continue

            if conditions:
                if not all(cond(request) for cond in conditions):
                    continue

            if excludes:
                exclusions = self.middleware_exclusions[excludes]
                if any(exc in executed_middlewares for exc in exclusions):
                    continue

            response = await middleware(request, response)
            executed_middlewares.add(middleware)

        return response
    
    async def handle_request(
        self,
        scope: Dict[str, Any],
        receive: Callable[..., Awaitable[Any]],
        send: Callable[..., Awaitable[None]],
    ) -> None:
        path = scope.get("path", "/")
        method = scope.get("method", "GET")
        request = Request(scope, receive)
        response = None

        try:
            allowed_methods = set()
            for (
                route_pattern,
                methods,
                handler,
                path_regex,
                response_model,
                endpoint,
            ) in self.routes:
                match = path_regex.match(path)
                if match:
                    if not methods or method in methods:
                        request.path_params = match.groupdict()

                        for transformer in self.request_transformers:
                            request = await transformer(request)

                        response = await handler(request)

                        if isinstance(response, str):
                            response = Response(content=response, content_type="text/plain")
                        elif isinstance(response, dict):
                            response = Response(content=json.dumps(response), content_type='application/json')
                        elif not isinstance(response, Response):
                            raise ValueError("Invalid response type")

                        if response_model:
                            response = self.validate_and_serialize_response(
                                response, response_model
                            )

                        break

                    else:
                        allowed_methods.update(methods)

            if response is None:
                if allowed_methods:
                    response = MethodNotAllowed(allowed_methods)
                else:
                    response = NotFound()

            for transformer in self.response_transformers:
                response = await transformer(response)

            response = await self.apply_middlewares(request, response)

            if self.compression_enabled:
                response = self.compress_response(response)

        except Exception as e:
            if isinstance(e, BadRequest):
                response = e
            elif isinstance(e, Forbidden):
                response = e
            elif isinstance(e, Unauthorized):
                response = e
            else:
                if self.debug:
                    response = await handle_exception(e)
                else:
                    response = InternalServerError()

        await response(scope, receive, send)

    def validate_and_serialize_response(
        self, response: Response, response_model: Type[T]
    ) -> Response:
        if not isinstance(response, response_model):
            raise ValueError(
                f"Response does not match the expected model {response_model.__name__}"
            )
        return Response(content=response.dict(), content_type="application/json")

    async def lifespan(
        self, scope: Dict[str, Any], receive: Callable[..., Awaitable[Any]], send: Callable[..., Awaitable[None]]
    ) -> None:
        started = False
        event = await receive() 
        try:
            if event['type'] == 'lifespan.startup':
                await self.startup()
                await send({"type": "lifespan.startup.complete"})
                started = True
                await receive()
        except BaseException:
            import traceback
            exc_text = traceback.format_exc()
            if started:
                await send({"type": "lifespan.shutdown.failed", "message": exc_text})
            else:
                await send({"type": "lifespan.startup.failed", "message": exc_text})
            raise
        else:
            await self.shutdown()
            await send({"type": "lifespan.shutdown.complete"})

    async def __call__(
        self, scope: Dict[str, str], receive, send
    ) -> None:
        if scope["type"] == "lifespan":
            await self.lifespan(scope, receive, send)
        elif scope["type"] == "http":
            try:
                await self.handle_request(scope, receive, send)
            except Exception as e:
                if self.debug:
                    response = await handle_exception(e)
                else:
                    response = InternalServerError()
                await response(scope, receive, send)

    async def startup(self):
        print('starting.....')

    async def shutdown(self):
        print('shutting down....')

    def generate_url(self, route_name: str, **kwargs: Any) -> str:
        if route_name in self.url_patterns:
            pattern = self.url_patterns[route_name]
            for key, value in kwargs.items():
                pattern = pattern.replace(f"{{{key}}}", str(value))
            return pattern
        else:
            raise ValueError(f"Route name '{route_name}' not found in url patterns.")

    def compress_response(self, response: Response) -> Response:
        if response.content_type and "text/" in response.content_type:
            compressed_content = zlib.compress(response.body.encode("utf-8"))
            response.body = compressed_content.decode("utf-8")
            response.headers["Content-Encoding"] = "gzip"
        return response



from aquilify.utils.http import parse_etags, parse_http_date_safe, quote_etag
from aquilify.utils._log import log_response
from aquilify.utils.regex_helper import _lazy_re_compile

from aquilify.wrappers import Response, Request

from typing import Any

from hashlib import md5

cc_delim_re = _lazy_re_compile(r"\s*,\s*")

def serialize_content(content: Any) -> bytes:
    if isinstance(content, (str, bytes, bytearray, memoryview)):
        return bytes(content, 'utf-8') if isinstance(content, str) else bytes(content)

    elif isinstance(content, (list, tuple, set, dict)):
        return str(content).encode('utf-8')

    else:
        raise TypeError(f"Unsupported content type: {type(content)}")

def set_response_etag(response: Response) -> Response:
    if not response.streaming and response.content:
        serialized_content = serialize_content(response.content)
        etag = md5(serialized_content).hexdigest()
        response.headers["ETag"] = quote_etag(etag)

    return response

def _precondition_failed(request):
    response = Response(status=412)
    log_response(
        "Precondition Failed: %s",
        request.path,
        response=response,
        request=request,
    )
    return response

def get_conditional_response(request: Request, etag=None, last_modified=None, response: Response=None):
    if response and not (200 <= response.status_code < 300):
        return response

    if_match_etags = parse_etags(request.headers.get("if-match", ""))
    if_unmodified_since = request.headers.get("if-unmodified-since")
    if_unmodified_since = if_unmodified_since and parse_http_date_safe(
        if_unmodified_since
    )
    if_none_match_etags = parse_etags(request.headers.get("if-none-match", ""))
    if_modified_since = request.headers.get("if-modified-since")
    if_modified_since = if_modified_since and parse_http_date_safe(if_modified_since)

    if if_match_etags and not _if_match_passes(etag, if_match_etags):
        return _precondition_failed(request)

    if (
        not if_match_etags
        and if_unmodified_since
        and not _if_unmodified_since_passes(last_modified, if_unmodified_since)
    ):
        return _precondition_failed(request)

    if if_none_match_etags and not _if_none_match_passes(etag, if_none_match_etags):
        if request.method in ("GET", "HEAD"):
            return response
        else:
            return _precondition_failed(request)
    if (
        not if_none_match_etags
        and if_modified_since
        and not _if_modified_since_passes(last_modified, if_modified_since)
        and request.method in ("GET", "HEAD")
    ):
        return response
    return response


def _if_match_passes(target_etag, etags):
    if not target_etag:
        return False
    elif etags == ["*"]:
        return True
    elif target_etag.startswith("W/"):
        return False
    else:
        return target_etag in etags


def _if_unmodified_since_passes(last_modified, if_unmodified_since):
    return last_modified and last_modified <= if_unmodified_since


def _if_none_match_passes(target_etag, etags):
    if not target_etag:
        return True
    elif etags == ["*"]:
        return False
    else:
        target_etag = target_etag.strip("W/")
        etags = (etag.strip("W/") for etag in etags)
        return target_etag not in etags


def _if_modified_since_passes(last_modified, if_modified_since):
    return not last_modified or last_modified > if_modified_since

def patch_vary_headers(response, newheaders):
    if response.headers.get("Vary"):
        vary_headers = cc_delim_re.split(response.headers["Vary"])
    else:
        vary_headers = []
    existing_headers = {header.lower() for header in vary_headers}
    additional_headers = [
        newheader
        for newheader in newheaders
        if newheader.lower() not in existing_headers
    ]
    vary_headers += additional_headers
    if "*" in vary_headers:
        response.headers["Vary"] = "*"
    else:
        response.headers["Vary"] = ", ".join(vary_headers)


import time
import hmac
import hashlib
import base64
import secrets
import string

import typing

from aquilify.settings import settings

class AxOTPError(Exception):
    pass

class AxOTP:
    HASH_FUNCTIONS = {
        'sha1': hashlib.sha1,
        'sha256': hashlib.sha256,
        'sha512': hashlib.sha512
    }

    def __init__(
        self, 
        secret_key: typing.Union[str, typing.Hashable] = None, 
        interval: int = 30, 
        digits: int = 6, 
        hash_func: typing.Callable[[], typing.Hashable] ='sha1', 
        counter: int = 0
    ) -> None:
        self.interval: int = getattr(settings.AXOTP_INTERVAL, 30) or interval
        self.digits: int = getattr(settings.AXOTP_DIGITS, 30) or digits
        self.counter: int = getattr(settings.AXOTP_COUNTER, 30) or counter
        self.hash_func: typing.Callable[[], typing.Hashable] = self._get_hash_func(hash_func)
        
        if secret_key:
            self.secret_key: typing.Union[str, typing.Hashable] = getattr(settings.AXOTP_SECRET_KEY, 30) or self._decode_secret(secret_key)
        else:
            self.secret_key: typing.Union[str, typing.Hashable] = self.gen_secret_key()

    def _get_hash_func(
        self, 
        hash_func_name
    ) -> typing.Callable[[], typing.Hashable]:
        hash_func = self.HASH_FUNCTIONS.get(hash_func_name.lower())
        if not hash_func:
            raise AxOTPError(f"Hash function '{hash_func_name}' not supported")
        return hash_func

    def _decode_secret(
        self,
        secret_key
    ) -> typing.Union[bytes, typing.ByteString]:
        try:
            return base64.b32decode(secret_key, casefold=True)
        except (ValueError, TypeError):
            raise AxOTPError("Invalid secret key format")

    def gen_secret_key(
        self,
        length=16
    ) -> typing.Union[bytes, typing.ByteString]:
        alphabet = string.ascii_uppercase + string.digits
        return ''.join(secrets.choice(alphabet) for _ in range(length)).encode()

    def gen_base32_secret_key(
        self,
        length=16
    ) -> typing.Union[str, typing.AnyStr]:
        return base64.b32encode(self.gen_secret_key(length)).decode()

    def _generate_mackey(
        self,
        counter
    ) -> typing.Any:
        counter_bytes = int.to_bytes(counter, length=8, byteorder='big')
        hash_result = hmac.new(self.secret_key, counter_bytes, self.hash_func).digest()
        offset = hash_result[-1] & 0x0F
        code = hash_result[offset:offset + 4]
        value = int.from_bytes(code, 'big') & 0x7FFFFFFF
        return value % (10 ** self.digits)

    def generate(
        self,
        timestamp=None
    ) -> typing.Any:
        timestamp = int(time.time() / self.interval) if timestamp is None else timestamp
        return self._generate_mackey(timestamp)

    def verify(
        self,
        otp_input
    ) -> typing.Any:
        return self.generate() == int(otp_input)

    def hash_funcs(self) -> typing.List[typing.Callable[..., typing.Hashable]]:
        return list(self.HASH_FUNCTIONS.keys())


import sqlite3


def dict_factory(cursor: sqlite3.Cursor) -> list[dict[str, object]]:

    """
    Converts a list of tuples to a list of dictionaries.

    :param cursor: SQLite database cursor
    :return: List of dictionaries
    """

    output = []
    for row in cursor.fetchall():
        output.append(
            {column[0]: row[index] for index, column in enumerate(cursor.description)}
        )

    return output



import importlib.util

from datetime import timedelta

class CSRFConfigSettings:
    def __init__(self, settings_module_path: str = "./settings.py") -> None:
        self.settings_module_path = settings_module_path
        self.csrf_data = []

    def fetch(self) -> None:
        try:
            with open(self.settings_module_path) as file:
                settings = self._load_settings_module()
                templates = getattr(settings, 'CSRF', [])
                self._extract_csrf(templates)

        except (FileNotFoundError, AttributeError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings

    def _extract_csrf(self, templates: list) -> None:
        for csrf in templates:
            extracted_data = self._extract_csrf_data(csrf)
            self.csrf_data.append(extracted_data)

    def _extract_csrf_data(self, csrf: dict) -> dict:
        backend = self._get_backend(csrf)
        options = self._get_options(csrf)

        return {
            'backend': backend,
            'options': {
                'secret_key': options.get('secret_key', ''),
                'expiration': options.get('expiration', timedelta(hours=1)),
                'token_key': options.get('token_key', "_csrf_token"),
                'logger': options.get('logger', None),
                'cookie_options': options.get('cookie_options', {"httponly": True, "secure": True, "samesite": 'strict'}),
                'protected_methods': options.get('protected_methods', ["POST", "PUT", "DELETE"]),
                'ip_verification': options.get('ip_verification', True),
                'trusted_ips': options.get('trusted_ips', ["*"]),
                'max_token_lifetime': options.get('max_token_lifetime', timedelta(hours=7)),
                'token_refresh_interval': options.get('token_refresh_interval', timedelta(hours=6)) 
            }
        }

    def _get_backend(self, csrf: dict) -> str:
        backend = csrf.get('BACKEND', None)
        return backend if backend else ""

    def _get_options(self, csrf: dict) -> dict:
        return csrf.get('OPTIONS', {})


import argparse
import logging
from aquilify.settings.base import settings
from aquilify.exception.base import ImproperlyConfigured
from aquilify.utils.module_loading import import_string
from aquilify.orm import connection

class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

class CommandChoices:
    MAKE_PORTING = 'makeporting'

def get_models_from_settings():
    try:
        database_settings = getattr(settings, "DATABASE")
        default_settings = database_settings.get('default', {})
        return default_settings.get('MODELS', [])
    except AttributeError:
        raise ImproperlyConfigured("Database settings not configured properly!")

def load_models(models):
    loaded_models = []
    for index, model in enumerate(models, start=1):
        loaded_model = import_string(model)
        if not loaded_model:
            raise ImproperlyConfigured(f"Model '{model}' not found")
        loaded_models.append(loaded_model)
        print(f"{Colors.OKGREEN}{index}.{Colors.ENDC} Model {Colors.OKGREEN}'{loaded_model.__name__}'{Colors.ENDC} initiated successfully... {Colors.OKGREEN}200{Colors.ENDC}")
    return loaded_models

def create_porting(models):
    try:
        print('')
        for index, model in enumerate(models, start=1):
            print(f"{Colors.OKGREEN}{index}.{Colors.ENDC} Model {Colors.OKGREEN}'{model.__name__}'{Colors.ENDC} Applied successfully... {Colors.OKGREEN}200{Colors.ENDC}")
        with connection(tables=models) as _:
            return lambda: connection(models)
    except Exception as e:
        raise ImproperlyConfigured(f"Error creating porting: {e}")

def setup_logging():
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    return logger

def make_porting(parser: argparse.ArgumentParser):
    logger = setup_logging()
    parser.add_argument('command', choices=[CommandChoices.MAKE_PORTING], help='Command to execute')
    args = parser.parse_args()

    if args.command == CommandChoices.MAKE_PORTING:
        try:
            models = get_models_from_settings()
            loaded_models = load_models(models)
            create_porting(loaded_models)
        except ImproperlyConfigured as e:
            logger.error(str(e))
            raise



import datetime
import functools
import os
import subprocess
import sys

from aquilify.utils.regex_helper import _lazy_re_compile

PY36 = sys.version_info >= (3, 6)
PY37 = sys.version_info >= (3, 7)
PY38 = sys.version_info >= (3, 8)
PY39 = sys.version_info >= (3, 9)
PY310 = sys.version_info >= (3, 10)
PY311 = sys.version_info >= (3, 11)
PY312 = sys.version_info >= (3, 12)


def get_version(version=None):
    """Return a PEP 440-compliant version number from VERSION."""
    version = get_complete_version(version)

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|rc}N - for alpha, beta, and rc releases

    main = get_main_version(version)

    sub = ""
    if version[3] == "alpha" and version[4] == 0:
        git_changeset = get_git_changeset()
        if git_changeset:
            sub = ".dev%s" % git_changeset

    elif version[3] != "final":
        mapping = {"alpha": "a", "beta": "b", "rc": "rc"}
        sub = mapping[version[3]] + str(version[4])

    return main + sub


def get_main_version(version=None):
    """Return main version (X.Y[.Z]) from VERSION."""
    version = get_complete_version(version)
    parts = 2 if version[2] == 0 else 3
    return ".".join(str(x) for x in version[:parts])


def get_complete_version(version=None):
    if version is None:
        from aquilify import VERSION as version
    else:
        assert len(version) == 5
        assert version[3] in ("alpha", "beta", "rc", "final")

    return version


def get_docs_version(version=None):
    version = get_complete_version(version)
    if version[3] != "final":
        return "dev"
    else:
        return "%d.%d" % version[:2]


@functools.lru_cache
def get_git_changeset():
    if "__file__" not in globals():
        return None
    repo_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    git_log = subprocess.run(
        "git log --pretty=format:%ct --quiet -1 HEAD",
        capture_output=True,
        shell=True,
        cwd=repo_dir,
        text=True,
    )
    timestamp = git_log.stdout
    tz = datetime.timezone.utc
    try:
        timestamp = datetime.datetime.fromtimestamp(int(timestamp), tz=tz)
    except ValueError:
        return None
    return timestamp.strftime("%Y%m%d%H%M%S")


version_component_re = _lazy_re_compile(r"(\d+|[a-z]+|\.)")


def get_version_tuple(version):
    """
    Return a tuple of version numbers (e.g. (1, 2, 3)) from the version
    string (e.g. '1.2.3').
    """
    version_numbers = []
    for item in version_component_re.split(version):
        if item and item != ".":
            try:
                component = int(item)
            except ValueError:
                break
            else:
                version_numbers.append(component)
    return tuple(version_numbers)



import importlib.util

class DispatcherConfigSettings:
    def __init__(self, settings_module_path="./settings.py") -> None:
        self.settings_module_path = settings_module_path
        self.dispatcher_instances = {}

    def fetch_dispatcher_config(self):
        try:
            settings = self._load_settings_module()
            dispatcher_config = getattr(settings, "APPLICATION_DISPATCHER", None)

            if dispatcher_config and len(dispatcher_config) == 2:
                root_instance, instance_dict = dispatcher_config

                if isinstance(root_instance, str) and isinstance(instance_dict, dict):
                    for key, value in instance_dict.items():
                        module = self._import_module(root_instance)
                        instance = self._get_instance_from_path(module, value)
                        if instance:
                            self.dispatcher_instances[key] = instance

        except (FileNotFoundError, AttributeError, ImportError) as e:
            raise ImportError(f"Error loading settings.py: {e}")

    def _load_settings_module(self):
        spec = importlib.util.spec_from_file_location("settings", self.settings_module_path)
        settings = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings)
        return settings

    def _import_module(self, module_name):
        return importlib.import_module(module_name)

    def _get_instance_from_path(self, module, instance_path):
        return getattr(module, instance_path) if instance_path else None



import importlib.util
import os
import sys

from .asgi_middleware  import ASGIMiddlewareLoader, ASGIMiddlewareLoaderError

from typing import Any, Dict, Optional, List

CachedModules = Dict[str, Any]

cached_modules: CachedModules = {}


class EntryPointExtractor:
    def __init__(self):
        pass

    def _extract_module_info(self, entry_point):
        module_info = {}
        if isinstance(entry_point, str):
            index_of_dot = entry_point.find('.')
            if index_of_dot != -1:
                module_info['module_path'] = entry_point[:index_of_dot]
                module_info['module_variable'] = entry_point[index_of_dot + 1:]
            else:
                module_info['module_path'] = entry_point
                module_info['module_variable'] = None
        return module_info

    def extractEntryPoint(self):
        current_directory = os.getcwd()
        settings_path = os.path.join(current_directory, 'settings.py')

        if os.path.exists(settings_path):
            spec = importlib.util.spec_from_file_location("settings", settings_path)
            settings = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(settings)

            if hasattr(settings, 'ENTRY_POINT'):
                entry_point = settings.ENTRY_POINT
                return self._extract_module_info(entry_point)
            else:
                return None
        else:
            return None

class ASGI:
    @staticmethod
    def wrap_with_middlewares(variable: Any, middlewares: List[Any]) -> Any:
        """Applies middlewares to the given variable."""
        for middleware in middlewares:
            variable = middleware(variable)
        return variable

    @staticmethod
    def load_module(module_name: str) -> Any:
        """Loads a Python module by name."""
        try:
            if module_name in sys.modules:
                return sys.modules[module_name]
            else:
                spec = importlib.util.find_spec(module_name)
                if spec is not None:
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)
                    sys.modules[module_name] = module
                    return module
                else:
                    raise ImportError(f"Module '{module_name}' not found in the Python path")
        except Exception as e:
            raise ImportError(f"Failed to load module '{module_name}': {e}")

    @staticmethod
    def get_variable_from_module(module: Any, variable_name: Optional[str]) -> Any:
        """Retrieves a variable from a loaded module."""
        if variable_name is None:
            return module
        else:
            imported_variable = getattr(module, variable_name, None)
            if imported_variable is not None:
                return imported_variable
            else:
                raise AttributeError(f"Variable '{variable_name}' not found in module '{module.__name__}'")

    @staticmethod
    def application() -> Any:
        """Loads an ASGI application."""
        entry_point = EntryPointExtractor().extractEntryPoint()
        module_name = entry_point['module_path'] or '__root__'
        variable_name = entry_point['module_variable'] or '__instance__'

        try:
            module = ASGI.load_module(module_name)
            imported_variable = ASGI.get_variable_from_module(module, variable_name)

            middleware_loader = ASGIMiddlewareLoader()
            middlewares = middleware_loader.load_asgi_middlewares()

            return ASGI.wrap_with_middlewares(imported_variable, middlewares)

        except ASGIMiddlewareLoaderError as e:
            raise ImportError(f"Failed to load ASGI middlewares: {e}")
        except ImportError as e:
            raise ImportError(f"Failed to import '{variable_name}' from '{module_name}': {e}")
        except Exception as e:
            raise ImportError(f"Failed to initialize ASGI application: {e}")


from functools import wraps

from aquilify.utils.functional import keep_lazy


class SafeData:
    __slots__ = ()

    def __html__(self):
        return self


class SafeString(str, SafeData):
    __slots__ = ()

    def __add__(self, rhs):
        t = super().__add__(rhs)
        if isinstance(rhs, SafeData):
            return SafeString(t)
        return t

    def __str__(self):
        return self


SafeText = SafeString 


def _safety_decorator(safety_marker, func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return safety_marker(func(*args, **kwargs))

    return wrapper


@keep_lazy(SafeString)
def mark_safe(s):
    if hasattr(s, "__html__"):
        return s
    if callable(s):
        return _safety_decorator(mark_safe, s)
    return SafeString(s)



import os
import secrets
import base64

from typing import Dict, Any
from datetime import datetime, timedelta

from aquilify.settings.sessions import SessionConfigSettings
from aquilify.settings import settings
from aquilify.utils.module_loading import import_string

_settings = SessionConfigSettings()

class InMemorySessionBeforeStage:
    def __init__(self) -> None:
        self.sessions: Dict[str] = {}
        self.session_lifetime: timedelta = timedelta(minutes=_settings.fetch().get('session_lifetime', 30))
        self.secret_key = _settings.fetch().get('secret_key', os.urandom(32))

    async def __call__(self, request: Any) -> Any:
        session_id = os.environ.get('X-SESSION_ID')
        if session_id is None or session_id not in self.sessions:
            session_id = self._generate_session_id()
            os.environ['X-SESSION_ID'] = session_id
            storage = None
            
            for item in settings.STORAGE_BACKEND:
                if "sessions" in item and "memory" in item["sessions"]:
                    storage = item["sessions"]["memory"]
                    break
            if not storage:
                raise ValueError("Either storage SESSION_BACKEND isn't found! or may not been configured properly!")
            func = import_string(storage)
            self.sessions[session_id] = func(session_id, self.secret_key)

        request.scope['session'] = self.sessions[session_id]
        self._cleanup_sessions()

    def _generate_session_id(self) -> str:
        return base64.urlsafe_b64encode(secrets.token_bytes(64)).decode('utf-8')

    def _cleanup_sessions(self) -> None:
        now = datetime.now()
        expired_sessions = [
            sid for sid, session in self.sessions.items()
            if now - session.updated_at > self.session_lifetime
        ]
        for sid in expired_sessions:
            self.sessions.pop(sid)

    def invalidate_session(self, session_id: str) -> None:
        if session_id in self.sessions:
            self.sessions.pop(session_id)
            os.environ.pop('X-SESSION_ID', None)



SCHEMATIC = """from aquilify.core.schematic import Schematic

# Schematic is a dynamic routing for Aquilify,
# You can use it part the routing in diffent octant :: @refer -> 3691

dynamicview = Schematic("my dynamic schematic") # Create a varible define the Schematic class and pass the name to the class.

# Define all you views...

async def home() -> dict:
    ...

# add the views here..

dynamicview.add_rule('/', methods=['GET', 'POST'], endpoint=home)

# don't forgot to link the schematic with you main application,
# In constructor.py using link function importing from aquilify.core.routing"""


from .wrappers import Response as BaseResponse

from typing import (
    Dict,
    Optional,
    Union,
    Callable,
    Any,
    List
)

from urllib.parse import (
    urlencode, 
    urlsplit, 
    urlunsplit, 
    quote
)

import os
import gzip
import io
import time
import json
import hashlib
import typing
import anyio

from datetime import datetime, timedelta
from mimetypes import MimeTypes
from functools import partial

from aquilify.utils.concurrency import iterate_in_threadpool
from aquilify.types import Scope, Receive, Send

class PlainTextResponse(BaseResponse):
    def __init__(
        self,
        content: Union[str, bytes, Callable, None] = None,
        status: Optional[int] = 200,
        headers: Optional[Dict[str, Union[str, int]]] = None,
        content_type: str = 'text/plain',
        encoding: Optional[str] = 'utf-8',
    ) -> None:
        """
        Create a plain text response.

        Args:
            content (Union[str, bytes, Callable, None]): The response content.
            status (Optional[int]): The HTTP status code (default is 200).
            headers (Optional[Dict[str, Union[str, int]]]): Additional headers for the response.
            content_type (str): The content type for the response (default is 'text/plain').
            encoding (Optional[str]): The character encoding for text content (default is 'utf-8').

        """
        super().__init__(content, status, headers)
        self.headers.setdefault('Content-Type', content_type)
        if encoding:
            self.headers['Content-Type'] += f'; charset={encoding}'

class JsonResponse(BaseResponse):
    def __init__(
        self,
        content: Union[Dict, Callable, None] = {},
        status: Optional[int] = 200,
        headers: Optional[Dict[str, Union[str, int]]] = None,
        content_type: str = 'application/json',
        encoding: Optional[str] = 'utf-8',
        validate: Optional[bool] = False,
    ) -> None:
        """
        Create a JSON response.

        Args:
            content (Union[Dict, Callable, None]): The response content (as a dictionary).
            status (Optional[int]): The HTTP status code (default is 200).
            headers (Optional[Dict[str, Union[str, int]]]): Additional headers for the response.
            content_type (str): The content type for the response (default is 'application/json').
            encoding (Optional[str]): The character encoding for JSON content (default is 'utf-8').
            validate (Optional[bool]): Whether to validate the JSON data (default is False).

        """
        if validate:
            try:
                json.dumps(content)
            except ValueError:
                raise ValueError("Invalid JSON content")

        super().__init__(json.dumps(content, ensure_ascii=False), status, headers)
        self.headers.setdefault('Content-Type', f'{content_type}; charset={encoding}')

class HTMLResponse(BaseResponse):
    def __init__(
        self,
        content: Union[str, bytes, Callable, None] = None,
        status: Optional[int] = 200,
        headers: Optional[Dict[str, Union[str, int]]] = None,
        content_type: str = 'text/html',
        encoding: Optional[str] = 'utf-8',
    ) -> None:
        """
        Create an HTML response.

        Args:
            content (Union[str, bytes, Callable, None]): The response content.
            status (Optional[int]): The HTTP status code (default is 200).
            headers (Optional[Dict[str, Union[str, int]]]): Additional headers for the response.
            content_type (str): The content type for the response (default is 'text/html').
            encoding (Optional[str]): The character encoding for text content (default is 'utf-8').

        """
        super().__init__(content, status, headers)
        self.headers.setdefault('Content-Type', f'{content_type}; charset={encoding}')
        
    def __str__(self):
        return f"HTMLResponse: Status={self.status_code}, Content-Type={self.content_type}, Content-Length={len(self.content)}"

class RedirectResponse(BaseResponse):
    def __init__(
        self,
        url: str,
        status_code: int = 307,
        headers: Optional[Dict[str, Union[str, int]]] = None,
        query_params: Optional[Dict[str, Union[str, int]]] = None,
        anchor: Optional[str] = None,
        delay: float = 0.0,
        content: Optional[str] = None,
    ) -> None:
        super().__init__(
            content=content or b"", status_code=status_code, headers=headers or {}
        )

        if delay > 0:
            time.sleep(delay)

        parsed_url = urlsplit(url)

        if query_params:
            parsed_url = self._add_query_params(parsed_url, query_params)

        if anchor:
            parsed_url = self._add_anchor(parsed_url, anchor)

        self.headers["Location"] = self._quote_url(parsed_url)

    def _add_query_params(
        self, parsed_url, query_params: Dict[str, Union[str, int]]
    ) -> str:
        query_dict = self._parse_query(parsed_url.query)
        query_dict.update(query_params)

        encoded_query = urlencode(query_dict, doseq=True)
        modified_url = parsed_url._replace(query=encoded_query)
        return modified_url

    def _add_anchor(self, parsed_url, anchor: str) -> str:
        modified_url = parsed_url._replace(fragment=quote(anchor))
        return modified_url

    def _parse_query(self, query: str) -> Dict[str, List[str]]:
        parsed_query = {}
        if query:
            for key, value in map(lambda x: x.split('=', 1), query.split('&')):
                parsed_query.setdefault(key, []).append(value)
        return parsed_query

    def _quote_url(self, parsed_url) -> str:
        return urlunsplit(
            (parsed_url.scheme, parsed_url.netloc, parsed_url.path, parsed_url.query, parsed_url.fragment)
        )
        
Content = typing.Union[str, bytes]
SyncContentStream = typing.Iterable[Content]
AsyncContentStream = typing.AsyncIterable[Content]
ContentStream = typing.Union[AsyncContentStream, SyncContentStream]

class StreamingResponse(BaseResponse):
    body_iterator: AsyncContentStream

    def __init__(
        self,
        content: ContentStream,
        status_code: int = 200,
        headers: typing.Optional[typing.Mapping[str, str]] = None,
        content_type: typing.Optional[str] = None,
    ) -> None:
        if isinstance(content, typing.AsyncIterable):
            self.body_iterator = content
        else:
            self.body_iterator = iterate_in_threadpool(content)
        self.status_code = status_code
        self.content_type = self.content_type if content_type is None else content_type
        
        super().__init__(content, status_code, headers, content_type)

    async def listen_for_disconnect(self, receive: Receive) -> None:
        while True:
            message = await receive()
            if message["type"] == "http.disconnect":
                break

    async def stream_response(self, send: Send) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.headers,
            }
        )
        async for chunk in self.body_iterator:
            if not isinstance(chunk, bytes):
                chunk = chunk.encode(self.encoding)
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async with anyio.create_task_group() as task_group:

            async def wrap(func: "typing.Callable[[], typing.Awaitable[None]]") -> None:
                await func()
                task_group.cancel_scope.cancel()

            task_group.start_soon(wrap, partial(self.stream_response, send))
            await wrap(partial(self.listen_for_disconnect, receive))
    

class FileResponse(BaseResponse):
    def __init__(
        self,
        file_path: str,
        filename: Optional[str] = None,
        status: Optional[int] = 200,
        headers: Optional[Dict[str, Union[str, int]]] = None,
        content_type: Optional[str] = None,
        inline: bool = False,
        buffer_size: int = 8192,
        chunked: bool = False,
        compress: bool = False,
        max_age: int = 3600,
        stream_file: bool = False,
        content_disposition: Optional[str] = None,
    ) -> None:
        super().__init__(content=None, status_code=status, headers=headers or {})
        self.buffer_size = buffer_size

        if not os.path.exists(file_path):
            raise FileNotFoundError("File not found")

        if not filename:
            filename = os.path.basename(file_path)

        if not content_type:
            content_type, _ = MimeTypes().guess_type(filename)

        disposition_type = content_disposition or ("inline" if inline else "attachment")
        self.headers["Content-Disposition"] = f'{disposition_type}; filename="{quote(filename)}'
        self.headers["Content-Type"] = content_type

        last_modified_time = os.path.getmtime(file_path)
        last_modified = datetime.utcfromtimestamp(last_modified_time).strftime('%a, %d %b %Y %H:%M:%S GMT')
        self.headers["Last-Modified"] = last_modified

        with open(file_path, "rb") as file:
            file_data = file.read()
            etag = hashlib.md5(file_data).hexdigest()
            self.headers["ETag"] = etag

        if stream_file:
            self._stream_file(file_path, buffer_size)
        else:
            if chunked:
                self._read_file_chunked(file_path, buffer_size)
            else:
                self._read_file_entirely(file_path)

        if compress:
            self._compress_content()

        self.enable_caching(max_age)

    def _stream_file(self, file_path: str, buffer_size: int) -> None:
        self.headers["Transfer-Encoding"] = "chunked"

        with open(file_path, "rb") as file:
            while True:
                chunk = file.read(buffer_size)
                if not chunk:
                    break
                chunk_size = hex(len(chunk))[2:]
                self.content += (f"{chunk_size}\r\n{chunk}\r\n").encode("utf-8")

        self.content += b"0\r\n\r\n"

    def _read_file_chunked(self, file_path: str, buffer_size: int) -> None:
        self.headers["Transfer-Encoding"] = "chunked"

        with open(file_path, "rb") as file:
            while True:
                chunk = file.read(buffer_size)
                if not chunk:
                    break
                chunk_size = hex(len(chunk))[2:]
                self.content += (f"{chunk_size}\r\n{chunk}\r\n").encode("utf-8")

        self.content += b"0\r\n\r\n"

    def _read_file_entirely(self, file_path: str) -> None:
        with open(file_path, "rb") as file:
            self.content = file.read()

        self.headers["Content-Length"] = str(len(self.content))

    def _compress_content(self) -> None:
        buffer = io.BytesIO()
        with gzip.GzipFile(fileobj=buffer, mode="wb") as f:
            f.write(self.content)
        compressed_content = buffer.getvalue()
        self.headers["Content-Encoding"] = "gzip"
        self.content = compressed_content

    def is_not_modified(self, request_headers: Dict[str, str], file_path: str) -> bool:
        if "If-None-Match" in request_headers and request_headers["If-None-Match"] == self.headers.get("ETag"):
            return True
        if "If-Modified-Since" in request_headers:
            try:
                client_timestamp = datetime.strptime(request_headers["If-Modified-Since"], '%a, %d %b %Y %H:%M:S GMT')
                file_timestamp = datetime.utcfromtimestamp(os.path.getmtime(file_path))
                if file_timestamp <= client_timestamp:
                    return True
            except ValueError:
                pass
        return False

    def handle_range_request(self, request_headers: Dict[str, str], file_path: str) -> Union[None, int]:
        if "Range" in request_headers:
            byte_range = request_headers["Range"]
            start, end = byte_range.strip("bytes=").split("-")
            start = int(start) if start else 0
            end = int(end) if end else os.path.getsize(file_path) - 1

            if start < 0 or end >= os.path.getsize(file_path) or start > end:
                return 416

            self.status_code = 206
            self.headers["Content-Range"] = f"bytes {start}-{end}/{os.path.getsize(file_path)}"
            self.headers["Content-Length"] = str(end - start + 1)
            self.content = self._stream_range(file_path, start, end, self.buffer_size)
            return None

    def _stream_range(self, file_path: str, start: int, end: int, buffer_size: int) -> Any:
        with open(file_path, "rb") as file:
            file.seek(start)
            while start <= end:
                chunk_size = min(end - start + 1, buffer_size)
                chunk = file.read(chunk_size)
                if not chunk:
                    break
                yield chunk
                start += len(chunk)

    def set_headers_for_conditional_request(self) -> None:
        self.headers["Date"] = format_date_time(datetime.utcnow())
        self.headers["Cache-Control"] = "public, max-age=0"

    def enable_caching(self, max_age: int = 3600) -> None:
        self.headers["Cache-Control"] = f"public, max-age={max_age}"
        expires = datetime.utcnow() + timedelta(seconds=max_age)
        self.headers["Expires"] = format_date_time(expires)

def format_date_time(dt: datetime) -> str:
    """
    Format a datetime object to the RFC 1123 date-time format.
    Args:
        dt (datetime): The datetime object to format.
    Returns:
        str: The formatted date-time string.
    """
    days = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
    months = [None, 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
    return f"{days[dt.weekday()]}, {dt.day:02d} {months[dt.month]} {dt.year} {dt.strftime('%H:%M:%S')} GMT"


from . import Request
from ..datastructure.core import MultiDict
from copy import deepcopy

from typing import (
    Any,
    Dict,
    Union,
    Optional,
    List,
    Callable
)

class ReqparserError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        return str(self.message)

    def __json__(self) -> str:
        return str(self)

    def __dict__(self) -> dict:
        return {'error': str(self)}

class Reqparser:
    def __init__(self) -> None:
        self.args: List[Dict[str, Union[str, bool, type, str, Any, Callable]]] = []

    def add_argument(
        self,
        name: str,
        required: bool = True,
        location: Union[str, List[str]] = 'json',
        type: type = str,
        help: str = '',
        default: Optional[Any] = None,
        store_missing: Optional[Any] = None,
        trim: bool = False,
        ignore: bool = False,
        choices: Optional[List[Any]] = None,
        error: Optional[str] = None,
        dest: Optional[str] = None,
        validation: Optional[Callable] = None
    ) -> None:
        arg_dict = MultiDict({
            'name': name,
            'required': required,
            'location': location if isinstance(location, list) else [location],
            'type': type,
            'help': help,
            'default': default,
            'store_missing': store_missing,
            'trim': trim,
            'ignore': ignore,
            'choices': choices,
            'error': error,
            'dest': dest if dest else name,
            'validation': validation
        })
        self.args.append(arg_dict)

    async def parse(self, request: Request) -> Dict[str, Any]:
        parsed_args: Dict[str, Any] = {}
        query_data: Dict[str, str] = request.args

        if request.method == 'GET':
            for arg in self.args:
                if 'query' in arg['location']:
                    value = query_data.get(arg['name'])
                    value = self._process_missing_value(arg, value)

                    if not arg['ignore']:
                        value = self._process_value(arg, value)

                    parsed_args[arg['dest']] = value
                else:
                    parsed_args[arg['dest']] = None

        else:
            json_data: Dict[str, Any] = await request.json()
            form_data: Dict[str, str] = await request.form()

            for arg in self.args:
                value = await self._get_value_by_location(request, arg, query_data, json_data, form_data)
                value = self._process_missing_value(arg, value)

                if not arg['ignore']:
                    value = self._process_value(arg, value)

                parsed_args[arg['dest']] = value

        return parsed_args

    async def _get_value_by_location(
        self,
        request: Request,
        arg: Dict[str, Any],
        query_data: Dict[str, str],
        json_data: Dict[str, Any],
        form_data: Dict[str, str]
    ) -> Any:
        for loc in arg['location']:
            if loc == 'query':
                value = query_data.get(arg['name'])
            elif loc == 'headers':
                value = request.headers.get(arg['name'])
            elif loc == 'form':
                value = form_data.get(arg['name'])
            elif loc == 'cookie':
                value = request.cookies.get(arg['name'])
            elif loc == 'json':
                value = json_data.get(arg['name'])
            else:
                raise ReqparserError(f"Invalid Location type: {loc}")

            if value is not None:
                return value

        return None

    def _process_missing_value(self, arg: Dict[str, Any], value: Any) -> Any:
        if value is None:
            if arg['required']:
                raise ReqparserError(arg['error'] if arg['error'] else f"Missing required argument: {arg['name']}. {arg['help']}")
            value = arg['default'] if arg['default'] is not None else arg['store_missing']
        return value

    def _process_value(self, arg: Dict[str, Any], value: Any) -> Any:
        if isinstance(value, str) and arg['trim']:
            value = value.strip()

        if arg['choices'] is not None and value not in arg['choices']:
            raise ReqparserError(arg['error'] if arg['error'] else f"Invalid value for argument '{arg['name']}'. Choose from: {arg['choices']}")

        if arg['type'] is not None:
            value = self._validate_and_convert_value(arg, value)

        return value

    def _validate_and_convert_value(self, arg: Dict[str, Any], value: Any) -> Any:
        if arg['type'] is int:
            try:
                value = int(value)
            except (ValueError, TypeError):
                raise ReqparserError(arg['error'] if arg['error'] else f"Invalid value for argument '{arg['name']}' type")
        elif arg['type'] is float:
            try:
                value = float(value)
            except (ValueError, TypeError):
                raise ReqparserError(arg['error'] if arg['error'] else f"Invalid value for argument '{arg['name']}' type")
            
        validation_function = arg.get('validation')
        if validation_function:
            try:
                validated_value = validation_function(value)
                value = validated_value
            except ValueError as ve:
                raise ReqparserError(str(ve))

        return value
    
    def copy(self) -> 'Reqparser':
        new_reqparser = Reqparser()
        new_reqparser.args = [arg_dict.copy() for arg_dict in self.args]
        return new_reqparser

    def replace_argument(self, name: str, new_argument: Dict[str, Any]) -> None:
        for arg_dict in self.args:
            if arg_dict['name'] == name:
                arg_dict.update(new_argument)

    def remove_argument(self, name: str) -> None:
        self.args = [arg_dict for arg_dict in self.args if arg_dict['name'] != name]



import re
import imghdr
import uuid
from urllib.parse import urlparse

from datetime import datetime
from typing import Any, Dict, List as ListType, Union

from .exceptions import (
    ValidationError,
    FieldValidationError
)

from .schema import Schema
from .schema import Field

class String(Field):
    def __init__(self, required: bool = False, min_length: int = None, max_length: int = None, regex: str = None, allow_blank: bool = False, strip_whitespace: bool = False, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.min_length = min_length
        self.max_length = max_length
        self.regex = regex
        self.allow_blank = allow_blank
        self.strip_whitespace = strip_whitespace

    async def validate(self, data: Any) -> str:
        if not isinstance(data, str):
            raise ValidationError(self.error_messages.get("validation", "Invalid data type. Expected string."))

        if self.strip_whitespace:
            data = data.strip()

        if not self.allow_blank and not data:
            raise ValidationError(self.error_messages.get("blank", "Field cannot be blank."))

        if self.min_length is not None and len(data) < self.min_length:
            raise ValidationError(f"Value must be at least {self.min_length} characters long.")

        if self.max_length is not None and len(data) > self.max_length:
            raise ValidationError(f"Value cannot exceed {self.max_length} characters.")

        if self.regex is not None and not re.match(self.regex, data):
            raise ValidationError(self.error_messages.get("invalid", "Invalid format."))

        return data

    async def serialize(self, data: str) -> str:
        if self.strip_whitespace:
            data = data.strip()

        if not self.allow_blank and not data:
            raise ValidationError(self.error_messages.get("blank", "Field cannot be blank."))

        return data

    async def serialize(self, data: str) -> str:
        if data is None and not self.allow_none:
            raise ValidationError(self.error_messages.get("null", "Field cannot be null."))
        return data


class Email(Field):
    USER_REGEX = r"(?=^.{1,64}$)^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+"
    DOMAIN_REGEX = r"(?=^.{1,253}$)(?:(?:\.{1}|\.{0,1}[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]+)*\.{1}[a-zA-Z]{2,})$"

    def __init__(self, required: bool = False, error_messages: Dict[str, str] = None, allow_name: bool = False, domain_whitelist: ListType[str] = None, domain_blacklist: ListType[str] = None):
        super().__init__(required, error_messages)
        self.allow_name = allow_name
        self.domain_whitelist = domain_whitelist or []
        self.domain_blacklist = domain_blacklist or []

    async def validate(self, data: Any) -> str:
        if not isinstance(data, str):
            raise ValidationError("Invalid data type. Expected string.")

        email_regex = r'^[\w\.-]+@[a-zA-Z\d\.-]+\.[a-zA-Z]{2,4}$'

        if not self.allow_name:
            email_regex = r'^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$'

        match = re.match(email_regex, data)
        if not match:
            raise ValidationError(self.error_messages.get("invalid", "Invalid email format."))

        if match:
            email = match.group()
            email_parts = email.split('@')
            email_local_part = email_parts[0]
            email_domain = email_parts[1]

            if not re.match(self.USER_REGEX, email_local_part):
                raise ValidationError(self.error_messages.get("local_part", "Invalid email local part."))

            if not re.match(self.DOMAIN_REGEX, email_domain):
                raise ValidationError(self.error_messages.get("domain", "Invalid email domain."))

            if self.domain_whitelist and email_domain not in self.domain_whitelist:
                raise ValidationError(self.error_messages.get("domain", "Email domain is not allowed."))

            if self.domain_blacklist and email_domain in self.domain_blacklist:
                raise ValidationError(self.error_messages.get("domain", "Email domain is blacklisted."))

        return data

    async def serialize(self, data: str) -> str:
        return data
    
class Integer(Field):
    def __init__(self, required: bool = False, allow_negative: bool = True, min_value: int = None, max_value: int = None, error_messages: Dict[str, str] = None, allow_divisible_by: int = None, allow_even: bool = False, allow_odd: bool = False):
        super().__init__(required, error_messages)
        self.allow_negative = allow_negative
        self.min_value = min_value
        self.max_value = max_value
        self.allow_divisible_by = allow_divisible_by
        self.allow_even = allow_even
        self.allow_odd = allow_odd

    async def validate(self, data: Any) -> int:
        if not isinstance(data, int):
            raise ValidationError("Invalid data type. Expected integer.")

        if not self.allow_negative and data < 0:
            raise ValidationError(self.error_messages.get("negative", "Value cannot be negative."))

        if self.min_value is not None and data < self.min_value:
            raise ValidationError(f"Value must be greater than or equal to {self.min_value}.")

        if self.max_value is not None and data > self.max_value:
            raise ValidationError(f"Value must be less than or equal to {self.max_value}.")

        if self.allow_even and data % 2 != 0:
            raise ValidationError("Value must be even.")

        if self.allow_odd and data % 2 == 0:
            raise ValidationError("Value must be odd.")

        if self.allow_divisible_by and data % self.allow_divisible_by != 0:
            raise ValidationError(f"Value must be divisible by {self.allow_divisible_by}.")

        return data

    async def serialize(self, data: int) -> int:
        return data

class Float(Field):
    def __init__(self, required: bool = False, min_value: float = None, max_value: float = None, allow_nan: bool = True, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.min_value = min_value
        self.max_value = max_value
        self.allow_nan = allow_nan

    async def validate(self, data: Any) -> float:
        if data is None and not self.allow_none:
            raise ValidationError(self.error_messages.get("null", "Field cannot be null."))

        if data is not None:
            try:
                float_data = float(data)
            except (ValueError, TypeError):
                raise ValidationError(self.error_messages.get("invalid", "Invalid data type. Expected float."))

            if not self.allow_nan and (float('nan') == float_data or float('-nan') == float_data):
                raise ValidationError(self.error_messages.get("nan", "Field does not allow NaN."))

            if self.min_value is not None or self.max_value is not None:
                if self.min_value is not None and self.max_value is not None:
                    if not self.min_value <= float_data <= self.max_value:
                        raise ValidationError(f"Value must be between {self.min_value} and {self.max_value}.")
                elif self.min_value is not None:
                    if float_data < self.min_value:
                        raise ValidationError(f"Value must be greater than or equal to {self.min_value}.")
                elif self.max_value is not None:
                    if float_data > self.max_value:
                        raise ValidationError(f"Value must be less than or equal to {self.max_value}")

        return float_data

    async def serialize(self, data: float) -> float:
        if data is None and not self.allow_none:
            raise ValidationError(self.error_messages.get("null", "Field cannot be null."))
        return data

class Date(Field):
    def __init__(self, required: bool = False, format: str = "%Y-%m-%d", min_date: str = None, max_date: str = None, allow_none: bool = False, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.format = format
        self.min_date = min_date
        self.max_date = max_date
        self.allow_none = allow_none

    async def validate(self, data: Any) -> str:
        if data is None and not self.allow_none:
            raise ValidationError(self.error_messages.get("null", "Field cannot be null."))

        if data is not None:
            try:
                parsed_date = datetime.strptime(data, self.format).date()
            except ValueError:
                raise ValidationError(self.error_messages.get("invalid", f"Invalid date format. Expected '{self.format}'."))

            if self.min_date is not None and parsed_date < datetime.strptime(self.min_date, self.format).date():
                raise ValidationError(f"Date should be after {self.min_date}.")
            
            if self.max_date is not None and parsed_date > datetime.strptime(self.max_date, self.format).date():
                raise ValidationError(f"Date should be before {self.max_date}.")

        return data

    async def serialize(self, data: str) -> str:
        if data is None and not self.allow_none:
            raise ValidationError(self.error_messages.get("null", "Field cannot be null."))
        return data
    
class NestedSchema(Field):
    def __init__(self, schema: Schema, required: bool = False, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.schema = schema

    async def validate(self, data: Any) -> Dict[str, Any]:
        if not isinstance(data, dict):
            raise ValidationError("Invalid data type. Expected dictionary.")

        try:
            return await self._validate_nested_schema(data)
        except ValidationError as e:
            raise ValidationError(f"Nested schema validation failed: {e.message}")

    async def _validate_nested_schema(self, data: Dict[str, Any]) -> Dict[str, Any]:
        if not data:
            if self.required:
                raise ValidationError("Nested data is required.")
            return {}

        validated_data = {}
        errors = {}
        
        for field_name, field_obj in self.schema.__dict__.items():
            if isinstance(field_obj, Field):
                field_data = data.get(field_name)
                if field_data is None and field_obj.required:
                    error_message = field_obj.error_messages.get("required", f"{field_name} is required.")
                    errors[field_name] = error_message
                if field_data is not None:
                    try:
                        validated_data[field_name] = await field_obj.validate(field_data)
                    except ValidationError as e:
                        errors[field_name] = str(e)

        if errors:
            raise FieldValidationError("Validation failed", errors)
        
        return validated_data

    async def serialize(self, data: Dict[str, Any]) -> Dict[str, Any]:
        return await self.schema.dumps(data)

class Bool(Field):
    def __init__(self, required: bool = False, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)

    async def validate(self, data: Any) -> bool:
        if isinstance(data, str):
            data = data.lower()
            if data in ['true', 't', 'yes', 'y', '1', 'false', 'f', 'no', 'n', '0']:
                return data in ['true', 't', 'yes', 'y', '1']
            else:
                raise ValidationError("Invalid boolean string representation.")

        if not isinstance(data, bool):
            raise ValidationError("Invalid data type. Expected boolean.")
        return data

    async def serialize(self, data: bool) -> bool:
        return data

class RegexMemorizer:
    _cache = {}

    @classmethod
    def compile(cls, pattern):
        if pattern not in cls._cache:
            cls._cache[pattern] = re.compile(pattern)
        return cls._cache[pattern]

class URL(Field):
    def __init__(self, required: bool = False, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.default_schemes = {'http', 'https', 'ftp', 'ftps'}
        self.scheme_re = RegexMemorizer.compile(r'^[a-z][a-z0-9+.-]*$')
        self.absolute_re = RegexMemorizer.compile(r'^/[^/\\]')
        self.relative_re = RegexMemorizer.compile(r'^[^/\\]')
        self.path_segment_re = RegexMemorizer.compile(r'^[^/\\]*$')

    async def validate(self, data: Any) -> str:
        if data is None:
            if self.required:
                raise ValidationError(self.error_messages.get("required", "Field is required."))
            return data

        if not isinstance(data, str):
            raise ValidationError("Invalid data type. Expected string.")

        parsed_url = urlparse(data)

        if not parsed_url.scheme or not parsed_url.netloc:
            raise ValidationError("Invalid URL format: Missing scheme or network location.")

        if parsed_url.scheme not in self.default_schemes:
            raise ValidationError("Invalid URL scheme. Only 'http', 'https', 'ftp', or 'ftps' schemes are allowed.")

        if not self.scheme_re.match(parsed_url.scheme):
            raise ValidationError("Invalid URL scheme format.")

        if parsed_url.port:
            if not (0 <= parsed_url.port <= 65535):
                raise ValidationError("Invalid URL port. Port must be between 0 and 65535.")

        domain_parts = parsed_url.netloc.split(':')
        domain = domain_parts[0]

        if not domain:
            raise ValidationError("Invalid URL format: Missing or invalid domain.")

        if domain != 'localhost':
            if len(domain.split('.')) <= 1 or len(domain.split('.')[-1]) < 2:
                raise ValidationError("Invalid URL format: Missing or invalid domain.")

        domain_name = domain.split('.')[-1]
        if domain_name != 'localhost' and not re.match(r'^[a-zA-Z0-9-]+$', domain_name):
            raise ValidationError("Invalid URL format: Domain contains invalid characters.")

        return data

    async def serialize(self, data: str) -> str:
        return data
    
class IPAddress(Field):
    def __init__(self, required: bool = False, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)

    async def validate(self, data: Any) -> str:
        if not isinstance(data, str):
            raise ValidationError("Invalid data type. Expected string.")

        ip_regex = r"^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
        if not re.match(ip_regex, data):
            raise ValidationError("Invalid IP address format.")
        return data

    async def serialize(self, data: str) -> str:
        return data
    
class Range(Field):
    def __init__(
        self,
        min: Union[int, float, None] = None,
        max: Union[int, float, None] = None,
        min_inclusive: bool = True,
        max_inclusive: bool = True,
        error_messages: Dict[str, str] = None,
    ):
        super().__init__(error_messages=error_messages)
        self.min_value = min
        self.max_value = max
        self.min_inclusive = min_inclusive
        self.max_inclusive = max_inclusive

    async def validate(self, data: Union[int, float]) -> Union[int, float]:
        error_messages = []

        if self.min_value is not None:
            if self.min_inclusive:
                if data < self.min_value:
                    error_messages.append(f"Value must be greater than or equal to {self.min_value}.")
            else:
                if data <= self.min_value:
                    error_messages.append(f"Value must be greater than {self.min_value}.")

        if self.max_value is not None:
            if self.max_inclusive:
                if data > self.max_value:
                    error_messages.append(f"Value must be less than or equal to {self.max_value}.")
            else:
                if data >= self.max_value:
                    error_messages.append(f"Value must be less than {self.max_value}.")

        if error_messages:
            raise ValidationError(", ".join(error_messages))

        return data

    async def serialize(self, data: Union[int, float]) -> Union[int, float]:
        return data

class Age(Field):
    def __init__(self, required: bool = False, min_age: int = None, max_age: int = None, age_range: Range = None, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.min_age = min_age
        self.max_age = max_age
        self.age_range = age_range

    async def validate(self, data: Any) -> int:
        if self.age_range:
            return await self.age_range.validate(data)

        if not isinstance(data, int):
            raise ValidationError("Age must be an integer.")

        if data < 0:
            raise ValidationError("Age cannot be negative.")

        if self.min_age is not None and data < self.min_age:
            raise ValidationError(f"Age should be at least {self.min_age} years.")

        if self.max_age is not None and data > self.max_age:
            raise ValidationError(f"Age cannot exceed {self.max_age} years.")

        return data

    async def serialize(self, data: int) -> int:
        return data
    
class Decimal(Field):
    def __init__(self, required: bool = False, min_value: float = None, max_value: float = None, precision: int = None, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.min_value = min_value
        self.max_value = max_value
        self.precision = precision

    async def validate(self, data: Any) -> float:
        if not isinstance(data, (int, float)):
            raise ValidationError(self.error_messages.get("invalid", "Invalid data type. Expected decimal or integer."))

        if isinstance(data, int):
            data = float(data)

        if self.min_value is not None and data < self.min_value:
            raise ValidationError(self.error_messages.get("min_value", f"Value must be greater than or equal to {self.min_value}."))

        if self.max_value is not None and data > self.max_value:
            raise ValidationError(self.error_messages.get("max_value", f"Value must be less than or equal to {self.max_value}"))

        if self.precision is not None:
            precision_str = str(data).split('.')
            if len(precision_str) > 1 and len(precision_str[1]) > self.precision:
                raise ValidationError(self.error_messages.get("precision", f"Value can have at most {self.precision} decimal places."))

        return data

    async def serialize(self, data: float) -> float:
        return data

class ListField(Field):
    def __init__(self, inner_field: Field, min_length: int = None, max_length: int = None, custom_validator=None, **kwargs):
        super().__init__(**kwargs)
        self.inner_field = inner_field
        self.min_length = min_length
        self.max_length = max_length
        self.custom_validator = custom_validator

    async def validate(self, data: Any) -> ListType:
        if not isinstance(data, list):
            raise ValidationError(self.error_messages.get("invalid", "Field must be a list."))

        if self.min_length is not None and len(data) < self.min_length:
            raise ValidationError(self.error_messages.get("min_length", f"List length should be at least {self.min_length}"))

        if self.max_length is not None and len(data) > self.max_length:
            raise ValidationError(self.error_messages.get("max_length", f"List length should be at most {self.max_length}"))

        if self.custom_validator:
            try:
                await self.custom_validator(data)
            except ValidationError as e:
                raise ValidationError(str(e))

        validated_data = []
        for index, item in enumerate(data):
            try:
                validated_item = await self.validate_item(item, index)
                validated_data.append(validated_item)
            except ValidationError as e:
                raise ValidationError({f"item_{index}": str(e)})
        return validated_data

    async def validate_item(self, item: Any, index: int) -> Any:
        try:
            validated_item = await self.inner_field.validate(item)
        except ValidationError as e:
            raise ValidationError({f"item_{index}": str(e)})
        return validated_item

    async def serialize(self, data: Any) -> ListType:
        if not isinstance(data, list):
            raise ValidationError(self.error_messages.get("invalid", "Field must be a list."))

        if self.min_length is not None and len(data) < self.min_length:
            raise ValidationError(self.error_messages.get("min_length", f"List length should be at least {self.min_length}"))

        if self.max_length is not None and len(data) > self.max_length:
            raise ValidationError(self.error_messages.get("max_length", f"List length should be at most {self.max_length}"))

        if self.custom_validator:
            try:
                await self.custom_validator(data)
            except ValidationError as e:
                raise ValidationError(str(e))

        serialized_data = []
        for index, item in enumerate(data):
            try:
                serialized_item = await self.serialize_item(item, index)
                serialized_data.append(serialized_item)
            except ValidationError as e:
                raise ValidationError({f"item_{index}": str(e)})
        return serialized_data

    async def serialize_item(self, item: Any, index: int) -> Any:
        try:
            serialized_item = await self.inner_field.serialize(item)
        except ValidationError as e:
            raise ValidationError({f"item_{index}": str(e)})
        return serialized_item

class UUID(Field):
    def __init__(self, required: bool = False, version: int = None, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.version = version

    async def validate(self, data: Any) -> str:
        if not isinstance(data, str):
            raise ValidationError(self.error_messages.get("invalid", "Invalid data type. Expected string."))

        try:
            uuid_obj = uuid.UUID(data)
        except ValueError:
            raise ValidationError(self.error_messages.get("format", "Invalid UUID format."))

        if self.version is not None and uuid_obj.version != self.version:
            raise ValidationError(self.error_messages.get("version", f"UUID should be version {self.version}."))

        return data

    async def serialize(self, data: str) -> str:
        return data


class File(Field):
    def __init__(self, required: bool = False, allowed_extensions: ListType[str] = None, max_size: int = None, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.allowed_extensions = allowed_extensions
        self.max_size = max_size

    async def validate(self, data: Any) -> Any:
        if not isinstance(data, bytes):
            raise ValidationError(self.error_messages.get("invalid", "Invalid file data."))

        if self.allowed_extensions:
            file_ext = imghdr.what(None, h=data)
            if file_ext not in self.allowed_extensions:
                raise ValidationError(self.error_messages.get("extensions", f"File extension not allowed. Allowed extensions: {', '.join(self.allowed_extensions)}"))

        if self.max_size is not None and len(data) > self.max_size:
            raise ValidationError(self.error_messages.get("max_size", f"File size exceeds the maximum allowed size: {self.max_size} bytes."))

        return data

    async def serialize(self, data: Any) -> Any:
        return data 
    
class Password(Field):
    def __init__(
        self,
        required: bool = False,
        min_length: int = 8,
        max_length: int = 64,
        special_chars: bool = True,
        uppercase: bool = True,
        lowercase: bool = True,
        numbers: bool = True,
        consecutive_chars: int = 3,
        error_messages: Dict[str, str] = None,
    ):
        super().__init__(required, error_messages)
        self.min_length = min_length
        self.max_length = max_length
        self.special_chars = special_chars
        self.uppercase = uppercase
        self.lowercase = lowercase
        self.numbers = numbers
        self.consecutive_chars = consecutive_chars

    async def validate(self, data: Any) -> str:
        if not isinstance(data, str):
            raise ValidationError("Invalid data type. Expected string.")

        if self.min_length > self.max_length:
            raise ValidationError("Minimum password length cannot be greater than maximum.")

        if len(data) < self.min_length:
            raise ValidationError(f"Password must be at least {self.min_length} characters long.")

        if len(data) > self.max_length:
            raise ValidationError(f"Password cannot exceed {self.max_length} characters.")

        if not self._check_password_complexity(data):
            raise ValidationError("Password complexity requirements not met.")

        if not self._check_consecutive_chars(data):
            raise ValidationError(f"Password should not contain {self.consecutive_chars} consecutive identical characters.")

        return data

    async def serialize(self, data: str) -> str:
        return data

    def _check_password_complexity(self, password: str) -> bool:
        has_special_chars = bool(re.search(r"[!@#$%^&*(),.?\":{}|<>]", password)) if self.special_chars else True
        has_uppercase = bool(re.search(r"[A-Z]", password)) if self.uppercase else True
        has_lowercase = bool(re.search(r"[a-z]", password)) if self.lowercase else True
        has_numbers = bool(re.search(r"\d", password)) if self.numbers else True

        return has_special_chars and has_uppercase and has_lowercase and has_numbers

    def _check_consecutive_chars(self, password: str) -> bool:
        for i in range(len(password) - self.consecutive_chars + 1):
            if len(set(password[i : i + self.consecutive_chars])) == 1:
                return False
        return True

class And(Field):
    def __init__(self, fields: ListType[Field], error_messages: Dict[str, str] = None):
        super().__init__(error_messages=error_messages)
        self.fields = fields

    async def validate(self, data: Dict[str, Any]) -> Dict[str, Any]:
        errors = {}

        for field in self.fields:
            try:
                validated_data = await field.validate(data)
                data.update(validated_data)
            except ValidationError as e:
                field_name = f"AND_{e.field_name}"
                errors[field_name] = e.message

        if errors:
            raise FieldValidationError("Validation failed", errors)

        return data

    async def serialize(self, data: Dict[str, Any]) -> Dict[str, Any]:
        serialized_data = {}
        errors = {}

        for field in self.fields:
            try:
                serialized_field = await field.serialize(data)
                serialized_data.update(serialized_field)
            except ValidationError as e:
                field_name = f"AND_{e.field_name}"
                errors[field_name] = e.message

        if errors:
            raise FieldValidationError("Serialization failed", errors)

        return serialized_data

class Length(Field):
    def __init__(self, equal: int = None, min_length: int = None, max_length: int = None, error_messages: Dict[str, str] = None):
        super().__init__(error_messages=error_messages)
        self.equal = equal
        self.min_length = min_length
        self.max_length = max_length

    async def validate(self, data: Any) -> Any:
        if data is None:
            raise ValidationError("Data is None.")

        try:
            length = len(data)
        except TypeError:
            raise ValidationError("Unable to determine the length of the data.")

        if self.equal is not None and length != self.equal:
            raise ValidationError(f"Length must be {self.equal} characters long.")
        if self.min_length is not None and length < self.min_length:
            raise ValidationError(f"Length must be at least {self.min_length} characters long.")
        if self.max_length is not None and length > self.max_length:
            raise ValidationError(f"Length cannot exceed {self.max_length} characters.")

        return data

    async def serialize(self, data: Any) -> Any:
        return data

class Regexp(Field):
    def __init__(self, regex: str, flags: int = 0, required: bool = False, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.regex = regex
        self.flags = flags
        self.compiled_regex = self.compile_regex(regex, flags)

    @staticmethod
    def compile_regex(pattern, flags):
        try:
            compiled = re.compile(pattern, flags)
        except re.error as e:
            raise ValueError(f"Invalid regular expression: {e}")
        return compiled

    async def validate(self, data: Any) -> str:
        if data is None:
            if self.required:
                raise ValidationError(self.error_messages.get("required", "Field is required."))
            return data

        if not isinstance(data, str):
            raise ValidationError(self.error_messages.get("invalid", "Value must be a string."))

        if not self.compiled_regex.match(data):
            error_message = self.error_messages.get("invalid", "Does not match the pattern.")
            raise ValidationError(error_message)

        return data

    async def serialize(self, data: str) -> str:
        return data

    def serialize_flags(self):
        return repr(re.ASCII if self.flags & re.ASCII else 0)
    
class Equal(Field):
    def __init__(self, equal_to: str, required: bool = False, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.equal_to = equal_to

    async def validate(self, data: Any) -> str:
        if data is None:
            if self.required:
                raise ValidationError(self.error_messages.get("required", "Field is required."))
            return data

        if data != self.equal_to:
            error_message = self.error_messages.get("invalid", f"Value must be equal to '{self.equal_to}'.")
            raise ValidationError(error_message)

        return data

    async def serialize(self, data: str) -> str:
        return data
    
class Dictonary(Field):
    def __init__(self, required: bool = False, min_items: int = None, max_items: int = None, schema: Schema = None, error_messages: Dict[str, str] = None):
        super().__init__(required, error_messages)
        self.min_items = min_items
        self.max_items = max_items
        self.schema = schema

    async def validate(self, data: Any) -> Dict[str, Any]:
        if not isinstance(data, dict):
            raise ValidationError("Data must be a dictionary.")

        if self.min_items is not None and len(data) < self.min_items:
            raise ValidationError(f"Dictionary must have at least {self.min_items} items.")

        if self.max_items is not None and len(data) > self.max_items:
            raise ValidationError(f"Dictionary cannot exceed {self.max_items} items.")

        if self.schema:
            validated_data = {}
            errors = {}
            for key, value in data.items():
                try:
                    validated_value = await self._validate_dict_item(value, key)
                    validated_data[key] = validated_value
                except FieldValidationError as e:
                    errors[key] = e.message
            if errors:
                raise FieldValidationError("Validation failed", errors)
            return validated_data

        return data

    async def _validate_dict_item(self, value: Any, key: str) -> Any:
        if not self.schema:
            return value

        try:
            return await self.schema.validate(value)
        except FieldValidationError as e:
            raise FieldValidationError(f"Error in field '{key}'", e.message)

    async def serialize(self, data: Dict[str, Any]) -> Dict[str, Any]:
        return data



def urlI8N():
    html_code = f'''
    <!doctype html>
    <html lang="en-us" dir="ltr">
        <head>
            <meta charset="utf-8">
            <title>The install worked successfully! Congratulations!</title>
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <style>
                .logo,h1{{font-size:1.375rem}}
                .option,footer{{display:grid;box-sizing:border-box}}
                .logo,.option{{text-decoration:none}}
                .option p,main p{{line-height:1.25}}
                a{{color:red}}
                header{{border-bottom:1px solid #efefef;display:grid;grid-template-columns:auto auto;align-items:self-end;justify-content:space-between;gap:7px;padding-top:20px;padding-bottom:10px}}
                body{{max-width:960px;color:#525252;font-family:"Segoe UI",system-ui,sans-serif;margin:0 auto}}
                main{{text-align:center}}
                h1,h2,h3,h4,h5,p,ul{{padding:0;margin:0;font-weight:400}}
                .logo{{font-weight:700}}
                .figure{{margin-top:19vh;max-width:265px;position:relative;z-index:-9;overflow:visible}}
                .exhaust__line{{animation:70ms ease-in-out 100 alternate thrust}}
                .smoke{{animation:.1s ease-in-out 70 alternate smoke}}
                @keyframes smoke{{0%{{transform:translate3d(-5px,0,0)}}100%{{transform:translate3d(5px,0,0)}}}}
                .flame{{animation:.1s ease-in-out 70 alternate burnInner2}}
                @keyframes burnInner2{{0%{{transform:translate3d(0,0,0)}}100%{{transform:translate3d(0,3px,0)}}}}
                @keyframes thrust{{0%{{opacity:1}}100%{{opacity:.5}}}}
                @media (prefers-reduced-motion:reduce){{.exhaust__line,.flame,.smoke{{animation:none}}}}
                h1{{max-width:32rem;margin:5px auto 0}}
                main p{{max-width:26rem;margin:15px auto 0}}
                footer{{grid-template-columns:1fr 1fr 1fr;gap:5px;padding:25px 0;position:fixed;left:50%;bottom:0;width:960px;transform:translateX(-50%);transform-style:preserve-3d;border-top:1px solid #efefef}}
                .option{{grid-template-columns:min-content 1fr;gap:10px}}
                .option svg{{width:1.5rem;height:1.5rem;fill:gray;border:1px solid #d6d6d6;padding:5px;border-radius:100%}}
                .option p{{font-weight:300;color:#525252;display:table}}
                .option .option__heading{{color:red;font-size:1.25rem;font-weight:400}}
                @media (max-width:996px){{body,footer{{max-width:780px}}}}
                @media (max-width:800px){{footer,header{{grid-template-columns:1fr}}footer{{height:100%;gap:60px;position:relative;padding:25px;width:100%;margin-top:50px}}.figure{{margin-top:10px}}main{{padding:0 25px}}main h1{{font-size:1.25rem}}header{{padding-left:20px;padding-right:20px}}}}
                @media (min-width:801px) and (max-height:730px){{.figure{{margin-top:80px}}}}
                @media (min-width:801px) and (max-height:600px){{footer{{position:relative;margin:135px auto 0}}.figure{{margin-top:50px}}}}
                .sr-only{{clip:rect(1px,1px,1px,1px);clip-path:inset(50%);height:1px;overflow:hidden;position:absolute;white-space:nowrap;width:1px}}
            </style>
        </head>
        <body>
        <header>
            <a class="logo" href="https://www.aquilify.vvfin.in/" target="_blank" rel="noopener">
                Aquilify
            </a>
            <p>View <a href="" target="_blank" rel="noopener">release notes</a> for Aquilify 1.13</p>
        </header>
        <main>
            <svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" version="1.1" width="640" height="480" viewBox="0 0 640 480" xml:space="preserve">
                <defs>
                </defs>
                <g transform="matrix(1 0 0 1 0 0)"  >
                <g style=""   >
                </g>
                </g>
                <g transform="matrix(1.43 0 0 1.43 307.97 231.91)"  >
                <g style=""   >
                        <g transform="matrix(1 0 0 1 0.02 0.02)"  >
                <path style="stroke: none; stroke-width: 1; stroke-dasharray: none; stroke-linecap: butt; stroke-dashoffset: 0; stroke-linejoin: miter; stroke-miterlimit: 4; fill: rgb(231,34,39); fill-rule: nonzero; opacity: 1;"  transform=" translate(-61.46, -61.46)" d="M 61.44 0 A 61.46 61.46 0 1 1 18 18 A 61.25 61.25 0 0 1 61.44 0 Z M 96.75 26.13 a 49.93 49.93 0 1 0 14.63 35.31 A 49.76 49.76 0 0 0 96.75 26.13 Z" stroke-linecap="round" />
                </g>
                        <g transform="matrix(1 0 0 1 -0.01 0)"  >
                <path style="stroke: none; stroke-width: 1; stroke-dasharray: none; stroke-linecap: butt; stroke-dashoffset: 0; stroke-linejoin: miter; stroke-miterlimit: 4; fill: rgb(38,38,38); fill-rule: nonzero; opacity: 1;"  transform=" translate(-61.43, -61.44)" d="M 69.6 51.38 h 11 A 2.9 2.9 0 0 1 83 55.89 L 56.78 100.67 a 2.89 2.89 0 0 1 -5.3 -2.07 l 3.87 -27.48 l -13.05 0.22 a 2.88 2.88 0 0 1 -2.93 -2.83 a 3 3 0 0 1 0.4 -1.51 L 65.59 22.23 a 2.89 2.89 0 0 1 5.34 1.87 L 69.6 51.38 Z" stroke-linecap="round" />
                </g>
                </g>
                </g>
                </svg>
            <h1>The install worked successfully! Congratulations!</h1>
            <p>You are seeing this page because <a href="" target="_blank" rel="noopener">DEBUG=True</a> is in your settings file and you have not configured any Constructors.</p>
        </main>
        </body>
    </html>
    '''
    return html_code



from __future__ import annotations

import sys
import traceback
import html
import typing
from typing import Callable, Type

if typing.TYPE_CHECKING:
    from ..wrappers import Response, Request

from .content import exceptions
from ._sys_info import get_sys_info

ExceptionHandlers = dict[Type[Exception], Callable]
exception_handlers: ExceptionHandlers = {}


def add_exception_handler(exception_cls: Type[Exception], handler: Callable):
    exception_handlers[exception_cls] = handler


def get_code_lines():
    exc_type, exc_value, exc_traceback = sys.exc_info()
    traceback_info = traceback.extract_tb(exc_traceback)
    relevant_code = []

    for filename, line_num, func_name, code_text in traceback_info:
        code_lines = []
        try:
            with open(filename) as file:
                code_lines = file.readlines()
        except FileNotFoundError as e:
            relevant_code.append(f"File '{filename}' not found: {str(e)}")
        except Exception as e:
            relevant_code.append(f"Error reading file '{filename}': {str(e)}")

        start_line = max(0, line_num - 5)
        end_line = min(len(code_lines), line_num + 5)

        relevant_code.append(f"In function '{func_name}' (File: '{filename}', Line: {line_num}):\n")

        max_digits = len(str(end_line))
        for i, line in enumerate(code_lines[start_line:end_line], start=start_line + 1):
            line = line.rstrip()
            marker = " >>> " if i == line_num else "     "
            line_num_str = str(i).rjust(max_digits)
            formatted_line = f"{marker}{line_num_str} | {line}"
            if i == line_num:
                formatted_line += f"   <-- Caused Exception: {exc_type.__name__}: {exc_value}"
            relevant_code.append(formatted_line)

        relevant_code.append('\n' + '-' * 60 + '\n')

        try:
            if hasattr(exc_value, 'args') and exc_value.args:
                relevant_code.append(f"Exception Arguments: {exc_value.args}\n")
        except Exception as e:
            relevant_code.append(f"Error retrieving exception arguments: {str(e)}\n")

        try:
            locals_dict = code_text and eval(code_text).f_locals
            if locals_dict:
                relevant_code.append("Local Variables:\n")
                for var_name, var_value in locals_dict.items():
                    if not var_name.startswith("__"):
                        relevant_code.append(f"{var_name}: {var_value}\n")
        except Exception as e:
            relevant_code.append(f"Error retrieving local variables: {str(e)}\n")

        try:
            globals_dict = code_text and eval(code_text).f_globals
            if globals_dict:
                relevant_code.append("Global Variables:\n")
                for var_name, var_value in globals_dict.items():
                    if not var_name.startswith("__"):
                        relevant_code.append(f"{var_name}: {var_value}\n")
        except Exception as e:
            relevant_code.append(f"Error retrieving global variables: {str(e)}\n")

    return '\n'.join(relevant_code)


async def default_exception_handler(error_message: str, traceback_info: str, code_lines, req_data = None) -> Response:
    from ..wrappers import Response
    try:
        formatted_traceback = "<br>".join(html.escape(line) for line in traceback_info.splitlines())
        error_type = traceback_info.splitlines()[-1].split(":")[0].strip()

        traceback_lines = traceback_info.splitlines()
        underlined_lines = 'No critical exception found by the handler!'
        for i, line in enumerate(traceback_lines):
            if "^" in line:
                underlined_line = traceback_lines[i - 1]
                underlined_lines = str(underlined_line)

        file_and_line = ""
        for line in traceback_lines:
            if "File " in line:
                file_and_line = line.strip()

        content = exceptions(error_message, formatted_traceback, underlined_lines, error_type, file_and_line,
                             code_lines, get_sys_info(), req_data)
        return Response(content, content_type='text/html', status_code=500)
    except Exception as e:
        error_message = f"Error in default exception handling: {str(e)}"
        return Response(content=error_message, content_type='text/plain', status_code=500)

def request_data(request: Request = None):
    if request is None:
        pass
    req_data = {
        "Host": request.host,
        "Port": request.url.port,
        "Client-IP": request.remote_addr,
        "Path": request.path,
        "Method": request.method,
        "User-Agent-String": request.user_agent,
        "Browser": request.user_agent.browser,
        "Timezone": request.user_agent.timezone,
        "Language": request.user_agent.language,
        "Platform": request.user_agent.platform,
        "Query-Params": request.query_params,
        "Base-Url": request.base_url,
        "Origin": request.origin,
        "Scheme": request.scheme,
        "Referer": request.referer,
        "Path-Params": request.path_param,
        "Session": request.session,
        "Cookie": request.cookies
    }
    return req_data

async def handle_exception(exception: Exception, request: Request = None) -> Response:
    try:
        traceback_info = traceback.format_exc()
        error_message = f"{str(exception)}"
        handler = exception_handlers.get(type(exception), default_exception_handler)
        req_data = request_data(request)
        return await handler(error_message, traceback_info, get_code_lines(), req_data)
    except Exception as e:
        error_message = f"{str(e)}"
        traceback_info = traceback.format_exc()
        return await default_exception_handler(error_message, traceback_info, get_code_lines())



from aquilify.wrappers import Request, Response
from aquilify.responses import HTMLResponse
from aquilify.settings import settings

class ConsoleAPI:
    def __init__(self, app) -> None:
        self.app = app
        self.prefix = getattr(settings, 'API_CONSOLE_URL', '/console')
        
    async def __call__(self, request: Request, response: Response):
        path: str = request.path
        self.methods = None

        if path.startswith(self.prefix):
            response = HTMLResponse(await self.html_data(request))
            return response
        return response
    
    async def html_data(self, request: Request):
        routes = self.app.routes

        javascript_code = """
            <script>
                async function fetchData(url, method, outputId, queryParams) {
                    const outputBox = document.getElementById(outputId);
                    outputBox.innerText = 'Loading...';
                    const headersBox = document.getElementById(outputId + '_headers');
                    headersBox.innerText = '';
                    const paramsBox = document.getElementById(outputId + '_params');
                    paramsBox.innerText = queryParams || 'No parameters';

                    try {
                        const response = await fetch(url, { method: method, timeout: 5000 });
                        const contentType = response.headers.get('content-type');
                        const headers = Object.fromEntries(response.headers);
                        const status = response.status;

                        headersBox.innerText = `HTTP Status: ${status}\n${JSON.stringify(headers, null, 2)}`;

                        let output = '';

                        if (contentType && contentType.includes('application/json')) {
                            output = await response.json();
                        } else {
                            output = await response.text();
                        }

                        outputBox.innerText = JSON.stringify(output, null, 2);
                        outputBox.style.display = 'block';
                        headersBox.style.display = 'block';
                        paramsBox.style.display = 'block';
                    } catch (error) {
                        if (error instanceof TypeError && error.message === 'Failed to fetch') {
                            outputBox.innerText = 'Error: Request timeout';
                        } else {
                            outputBox.innerText = `Error: ${error.message}`;
                        }
                        outputBox.style.display = 'block';
                    }
                }

                function clearResponseBoxes() {
                    const responseBoxes = document.querySelectorAll('.response-box');
                    responseBoxes.forEach(box => {
                        box.innerText = '';
                        box.style.display = 'none';
                    });

                    const headersBoxes = document.querySelectorAll('.headers-box');
                    headersBoxes.forEach(box => {
                        box.innerText = '';
                        box.style.display = 'none';
                    });

                    const paramsBoxes = document.querySelectorAll('.params-box');
                    paramsBoxes.forEach(box => {
                        box.innerText = 'No parameters';
                        box.style.display = 'none';
                    });
                }

                function changeMethod(url, method, outputId, contentType) {
                    const fetchButton = document.getElementById(`fetchButton_${outputId}`);
                    fetchButton.onclick = function() {
                        const payload = document.getElementById(`${outputId}_payload`).value;
                        fetchData(url, method, outputId, '{}', contentType, payload);
                    };
                }


                function expandCollapseAll(action) {
                    const headersBoxes = document.querySelectorAll('.headers-box');
                    headersBoxes.forEach(box => {
                        box.style.display = action;
                    });

                    const paramsBoxes = document.querySelectorAll('.params-box');
                    paramsBoxes.forEach(box => {
                        box.style.display = action;
                    });
                }
            </script>
            <style>
                body {
                    font-family: 'Segoe UI', Tahoma, Geneva, Verdana, sans-serif;
                    padding: 5px;
                    margin: 0;
                    background-color: #f5f5f5;
                    color: #333;
                    line-height: 1.6;
                }

                h2 {
                    text-align: center;
                    color: #444;
                    margin-bottom: 20px;
                }

                button {
                    padding: 10px 20px;
                    margin: 5px;
                    cursor: pointer;
                    background-color: rgb(7, 170, 235);
                    color: #fff;
                    border: none;
                    border-radius: 4px;
                    transition: background-color 0.3s;
                }

                button:hover {
                    background-color: #2980b9;
                }

                select {
                    padding: 8px;
                }

                .response-box, .headers-box, .params-box {
                    white-space: pre-wrap;
                    font-family: 'Courier New', monospace;
                    color: white;
                    padding: 10px;
                    margin-bottom: 10px;
                    background-color: black;
                    border-radius: 4px;
                    overflow: auto;
                    display: none;
                }

                .toggle-headers, .toggle-params {
                    cursor: pointer;
                    color: #3498db;
                    text-decoration: underline;
                    margin-left: 10px;
                }
            </style>
        """

        html_data = """
            <!DOCTYPE html>
            <html lang="en">
            <head>
                <meta charset="UTF-8">
                <meta name="viewport" content="width=device-width, initial-scale=1.0">
                <title>ConsoleAPI | AQUILIFY - v1.1</title>
            </head>
            <body>
                <h2>Welcome to AQUILIFY ConsoleAPI - v1.1</h2>
                <button onclick="expandCollapseAll('block')">Expand All</button>
                <button onclick="expandCollapseAll('none')">Collapse All</button>
                <h3>Available Routes:</h3>
        """

        for index, route in enumerate(routes):
            path, methods, _, _, _, _ = route
            path = self.convert_regex_path(path)
            output_id = f"response_{index}"  # Unique ID for each output box
            
            # HTML for each route container
            route_html = f"""
                <div style="background-color: #f5f5f5; padding: 20px; margin-bottom: 10px;">
                    <h4>PATH - '{path}' & URL - <u>{request.scheme}://{request.host}{path}</u></h4>
                    <div>
                        <label for="method_{index}">Method:</label>
                        <select id="method_{index}" onchange="changeMethod('{request.scheme}://{request.host}{path}', this.value, '{output_id}')">
            """

            for method in methods:
                route_html += f"<option value='{method}'>{method}</option>"

            route_html += "</select></div>"
            route_html += f"""
                <div>
                    <button onclick="fetchData('{request.scheme}://{request.host}{path}', document.getElementById('method_{index}').value, '{output_id}', '{str(request.query_params)}')">Fetch</button>
                    <span class="toggle-headers" onclick="document.getElementById('{output_id + '_headers'}').style.display = (document.getElementById('{output_id + '_headers'}').style.display === 'block' ? 'none' : 'block')">Headers</span>
                    <span class="toggle-params" onclick="document.getElementById('{output_id + '_params'}').style.display = (document.getElementById('{output_id + '_params'}').style.display === 'block' ? 'none' : 'block')">Params</span>
                </div>
                <div id="{output_id}" class="response-box"></div>
                <div id="{output_id + '_headers'}" class="headers-box"></div>
                <div>
                    <div id="{output_id + '_params'}" class="params-box">Query Params: {str(request.query_params)}</div>
                </div>
            """

            route_html += "</div>"  # Close the route container
            html_data += route_html  # Append the route HTML to the main HTML

        # Finish the main HTML data
        html_data += """
                <button onclick="clearResponseBoxes()">Clear All Responses</button>
            </body>
            </html>
        """

        final_data = f"{javascript_code}{html_data}"  # Combine JavaScript and HTML
        return final_data

    def convert_regex_path(self, path):
        return path[1:-1] if path.startswith('^') and path.endswith('$') else path


# AQUILIFY BASE MIDDLEWARE
# refer @noql -> 5806

# Aquilify base middleware is still under testing, it will be released very soon...

# For more info visit :: http://aquilify.vvfin.in/middlewares/base/#main


from .cookie import CookieSessionStorage as CookieSessionStorage
from .memory import InMemorySessionStorage as InMemorySessionStorage

__all__ = [
    CookieSessionStorage,
    InMemorySessionStorage
]